"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
"""
from collections.abc import Mapping, Sequence
from collections.abc import Mapping as Mapping_Type
from typing import Annotated, Any, Literal, Union

from pydantic import BaseModel, BeforeValidator, ConfigDict, Field, ImportString, StringConstraints, field_validator

from corl.libraries.factory import Factory
from corl.libraries.parameters import Parameter
from corl.libraries.units import Quantity, corl_get_ureg
from corl.libraries.utils import replace_magic_strings
from corl.simulators.base_platform import BasePlatform

ObjectStoreElem = Annotated[Parameter | Quantity | Any, BeforeValidator(Factory.resolve_factory)]


class Functor(BaseModel):
    """
    - name: The optional name of the functor.
    - class: The class of the functor.
    - config: The functor's specific configuration dictionary.
    """

    functor: ImportString
    name: Annotated[str, StringConstraints(strip_whitespace=True, min_length=1)] = Field(validate_default=True, default="")
    config: dict[str, ObjectStoreElem] = {}
    references: dict[str, str] = {}
    model_config = ConfigDict(arbitrary_types_allowed=True, extra="forbid")

    @field_validator("name", mode="before")
    @classmethod
    def name_from_functor(cls, v, values):
        """Create default value for name from the functor name"""
        # "'functor' not in values" means ImportString failed and an error is coming
        # Check for it here because that error is more helpful than a KeyError here.
        if not v and "functor" in values.data:
            return values.data["functor"].__name__
        return v

    def create_functor_object(
        self, param_sources: Sequence[Mapping[str, Any]] = (), ref_sources: Sequence[Mapping[str, Any]] = (), **kwargs
    ):
        """Create the object with this functor

        TODO:  Better description
        """
        functor_args = self.resolve_storage_and_references(
            param_sources=param_sources,
            ref_sources=ref_sources,
            platform=kwargs.get("platform", None),
        )
        return self.functor(name=self.name, **functor_args, **kwargs)

    def resolve_to_quantity(self, value):
        """
        recurse into the configuration of value and convert all Quantity patterns to Quantities
        """
        if isinstance(value, dict):
            if "value" in value and ("units" in value or "unit" in value) and len(value.keys()) == 2:
                return corl_get_ureg().Quantity(value=value["value"], units=value.get("units", value.get("unit")))
            for key, val in value.items():
                value[key] = self.resolve_to_quantity(val)
        if isinstance(value, list):
            value = [self.resolve_to_quantity(x) for x in value]
        return value

    def resolve_storage_and_references(
        self,
        param_sources: Sequence[Mapping[str, Any]] = (),
        ref_sources: Sequence[Mapping[str, Any]] = (),
        platform: BasePlatform | None = None,
    ):
        """Resolve parameter storage and references to get direct functor arguments."""

        functor_args: dict[str, Any] = {}
        functor_units = getattr(self.functor, "REQUIRED_UNITS", {})

        # note, I don't think this ever worked as intended, and now that we have pint
        # and units everywhere the pathways it goes down cause issues
        # config_units = self.config.get('unit', self.config.get("units", None))
        # if config_units is not None and not isinstance(config_units, str):
        #     raise TypeError(f'Units of {self.name} are "{config_units}", which is not a string')

        for arg_name, arg_value in self.config.items():
            new_val = self.resolve_to_quantity(arg_value)

            # Resolve parameter values
            if isinstance(new_val, Parameter):
                for source in param_sources:
                    if arg_name in source.get(self.name, {}):
                        resolved_value = source[self.name][arg_name]
                        break
                else:
                    # This "else" means "no break encountered", which means that the argument was not found in any source
                    raise RuntimeError(f"Could not resolve argument {arg_name} for {self.name} from the parameter sources")
            elif isinstance(arg_value, dict) and "value" in arg_value and "units" in arg_value:
                # {'value: 0, "units: 'meter'} - > Quantity(0, 'meter')
                resolved_value = corl_get_ureg().Quantity(value=arg_value["value"], units=arg_value["units"])
            else:
                resolved_value = new_val

            # Resolve units
            functor_args[arg_name] = self._resolve_units(
                name=arg_name, value=resolved_value, functor_units=functor_units, error_name=self.name
            )

        for ref_dest, ref_src in self.references.items():
            true_dest = ref_dest.split(".")
            true_ref_src = ref_src
            if platform is not None:
                true_ref_src = replace_magic_strings(true_ref_src, platform_name=platform.name)

            # Resolve references
            for source in ref_sources:
                if true_ref_src in source:
                    ref_obj = source[true_ref_src]
                    if not isinstance(ref_obj, Parameter):
                        break
            else:
                # This "else" means "no break encountered", which means either:
                # 1. true_ref_src was not found in any source
                # 2. true_ref_src was found; however, it was an unresolved Parameter
                raise RuntimeError(f"Could not find {ref_src} -> {true_ref_src} for {self.name} in the reference storage")

            # Resolve units
            tmp_functor_args = functor_args
            for functor_arg_path in true_dest[:-1]:
                tmp_functor_args = tmp_functor_args.setdefault(functor_arg_path, {})
            tmp_functor_args[true_dest[-1]] = self._resolve_units(
                name=ref_dest, value=ref_obj, functor_units=functor_units, error_name=self.name
            )

        return functor_args

    def add_to_parameter_store(self, parameter_store: dict[str, dict[str, Parameter]]) -> None:
        """Add the parameters of this functor to an external parameter store.

        Parameters
        ----------
        parameter_store : Dict[str, Dict[str, Parameter]]
            Parameter store to which to add the parameters.  The keys of the outer dictionary are functor names.  The inner dictionary is
            the collection of Parameters.
        """
        parameters = {k: v for k, v in self.config.items() if isinstance(v, Parameter)}
        if parameters:
            if self.name in parameter_store:
                raise ValueError(f"Duplicate functor name in parameter store: {self.name}")
            parameter_store[self.name] = parameters

    @staticmethod
    def _resolve_units(name: str, value: ObjectStoreElem, functor_units: dict[str, str | Literal[True]], error_name: str) -> Any:
        assert not isinstance(value, Parameter)

        functor_arg: Any

        # Determine what to pass in to functor
        if isinstance(value, Quantity):
            # Value has units
            if name in functor_units:
                # Extra local variable required so that MyPy does proper type reduction in the conditional block below.
                these_units = functor_units[name]
                # Require it to be actually boolean True, not just "truthy"
                functor_arg = value if these_units is True else value.to(these_units)
            else:
                functor_arg = value
        else:
            # Value has no units
            # Functor requires them and the unit is not None
            if (not isinstance(value, str)) and functor_units.get(name, None) is not None:
                raise RuntimeError(f"Argument {name} of {error_name} is missing required units")

            # Functor does not require them
            functor_arg = value

        return functor_arg


class FunctorWrapper(Functor):
    """
    wrapped: The functor or functor wrapper configuration wrapped by this functor wrapper.
    """

    wrapped: Union["FunctorMultiWrapper", "FunctorWrapper", "FunctorDictWrapper", Functor]

    def create_functor_object(
        self, param_sources: Sequence[Mapping[str, Any]] = (), ref_sources: Sequence[Mapping[str, Any]] = (), **kwargs
    ):
        """Create the object with this functor

        TODO:  Better description
        TODO:  Clean up logic so does not have too-many-branches
        """

        wrapped_func = self.wrapped.create_functor_object(param_sources=param_sources, ref_sources=ref_sources, **kwargs)

        functor_args = self.resolve_storage_and_references(
            param_sources=param_sources,
            ref_sources=ref_sources,
            platform=kwargs.get("platform", None),
        )

        return self.functor(name=self.name, wrapped=wrapped_func, **functor_args, **kwargs)

    def add_to_parameter_store(self, parameter_store: dict[str, dict[str, Parameter]]) -> None:
        """Add the parameters of this functor to an external parameter store.

        Parameters
        ----------
        parameter_store : Dict[str, Dict[str, Parameter]]
            Parameter store to which to add the parameters.  The keys of the outer dictionary are functor names.  The inner dictionary is
            the collection of Parameters.
        """
        super().add_to_parameter_store(parameter_store)
        self.wrapped.add_to_parameter_store(parameter_store)


class FunctorMultiWrapper(Functor):
    """
    wrapped: The functor or functor wrapper configuration wrapped by this functor wrapper.
    """

    wrapped: list[Union["FunctorMultiWrapper", FunctorWrapper, Functor, "FunctorDictWrapper"]]

    def create_functor_object(
        self, param_sources: Sequence[Mapping[str, Any]] = (), ref_sources: Sequence[Mapping[str, Any]] = (), **kwargs
    ):
        """Create the object with this functor

        TODO:  Better description
        TODO:  Clean up logic so does not have too-many-branches
        """
        wrapped_funcs = [x.create_functor_object(param_sources=param_sources, ref_sources=ref_sources, **kwargs) for x in self.wrapped]

        functor_args = self.resolve_storage_and_references(
            param_sources=param_sources,
            ref_sources=ref_sources,
            platform=kwargs.get("platform", None),
        )

        return self.functor(name=self.name, wrapped=wrapped_funcs, **functor_args, **kwargs)

    def add_to_parameter_store(self, parameter_store: dict[str, dict[str, Parameter]]) -> None:
        """Add the parameters of this functor to an external parameter store.

        Parameters
        ----------
        parameter_store : Dict[str, Dict[str, Parameter]]
            Parameter store to which to add the parameters.  The keys of the outer dictionary are functor names.  The inner dictionary is
            the collection of Parameters.
        """
        super().add_to_parameter_store(parameter_store)
        for wrapped in self.wrapped:
            wrapped.add_to_parameter_store(parameter_store)


class FunctorDictWrapper(Functor):
    """
    wrapped: The dict of functor or functor wrapper configurations wrapped by this wrapper
    """

    wrapped: dict[str, Union["FunctorDictWrapper", FunctorMultiWrapper, FunctorWrapper, Functor]]

    @field_validator("wrapped", mode="before")
    @classmethod
    def adjust_wrapped_name(cls, v):
        """Use the dictionary wrapping key as the name if none is provided."""
        if not isinstance(v, Mapping_Type):
            return v
        # this pre validator is pretty dangerous in terms of
        # what it will attempt to run on, because a regular wrapper
        # has a dictionary it will attempt to perform this operation
        # on regular wrappers, so we are just going to verify that
        # this is not a regular wrapper by checking for a functor appearing
        # in wrapped
        if "functor" in v:
            return v
        for key, value in v.items():
            try:
                if "name" not in value:
                    value["name"] = key
            except:  # noqa: E722,S110,PERF203
                pass
        return v

    def create_functor_object(
        self, param_sources: Sequence[Mapping[str, Any]] = (), ref_sources: Sequence[Mapping[str, Any]] = (), **kwargs
    ):
        """Create the object with this functor

        TODO:  Better description
        TODO:  Clean up logic so does not have too-many-branches
        """
        wrapped_funcs = {
            k: v.create_functor_object(param_sources=param_sources, ref_sources=ref_sources, **kwargs) for k, v in self.wrapped.items()
        }

        functor_args = self.resolve_storage_and_references(
            param_sources=param_sources,
            ref_sources=ref_sources,
            platform=kwargs.get("platform", None),
        )

        return self.functor(name=self.name, wrapped=wrapped_funcs, **functor_args, **kwargs)

    def add_to_parameter_store(self, parameter_store: dict[str, dict[str, Parameter]]) -> None:
        """Add the parameters of this functor to an external parameter store.

        Parameters
        ----------
        parameter_store : Dict[str, Dict[str, Parameter]]
            Parameter store to which to add the parameters.  The keys of the outer dictionary are functor names.  The inner dictionary is
            the collection of Parameters.
        """
        super().add_to_parameter_store(parameter_store)
        for wrapped in self.wrapped.values():
            wrapped.add_to_parameter_store(parameter_store)


FunctorWrapper.update_forward_refs()
FunctorMultiWrapper.update_forward_refs()
FunctorDictWrapper.update_forward_refs()



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
Defines a task
"""
import pathlib
import warnings

from pydantic import BaseModel, ConfigDict, Field, dataclasses, field_validator, model_validator

from corl.environment.default_env_rllib_callbacks import DefaultCallbacks, EnvironmentDefaultCallbacks
from corl.evaluation.runners.section_factories.teams import Teams
from corl.experiments.base_experiment import ExperimentParse
from corl.experiments.rllib_utils.policy_mapping_functions import PolicyIsAgent
from corl.parsers.yaml_loader import load_file


class Config:
    """Pydantic options"""

    arbitrary_types_allowed = True


@dataclasses.dataclass(config=Config)
class Task:
    """Rllib task to be performed"""

    config_yaml_file: str | pathlib.Path | None
    environment: str = "default"
    experiment_parse: ExperimentParse = Field(default=None, validate_default=True)  # type: ignore

    @field_validator("config_yaml_file")
    @classmethod
    def validate_config_yaml_file(cls, v):
        """Ensure config_yaml_file is a pathlib.Path"""
        return pathlib.Path(v)

    @field_validator("experiment_parse", mode="before")
    def validate_experiment_parse(cls, v, values):
        """load experiment_parse"""
        if v is None:
            assert "config_yaml_file" in values.data
            assert "environment" in values.data
            experiment_config = load_file(config_filename=str(values.data["config_yaml_file"]))
            env_config = experiment_config["config"]["env_config"]
            if isinstance(env_config, dict):
                if values.data["environment"] in env_config:
                    env_config = env_config[values.data["environment"]]
                if isinstance(env_config, list):
                    warnings.warn("Applying patches without the !merge yaml directive is deprecated", DeprecationWarning)
                elif "simulator" not in env_config or "episode_parameter_provider" not in env_config:
                    raise ValueError(f"Invalid experiment config for environment {values.data['environment']}")
                experiment_config["config"]["env_config"] = env_config

            return ExperimentParse(**experiment_config)

        assert isinstance(v, ExperimentParse), "experiment_parse was provided but was not an instance of ExperimentParse"
        return v


class Experiment(BaseModel):
    """configs for running an experiment of the provided task/teams"""

    model_config = ConfigDict(arbitrary_types_allowed=True)

    task: Task
    teams: Teams

    rllib_config: dict
    callbacks: type[DefaultCallbacks] = EnvironmentDefaultCallbacks
    tune_config: dict
    ray_config: dict
    env_config: dict

    @model_validator(mode="before")
    @classmethod
    def validate_experiment(cls, values):
        """Initializes experiment configs using the provided task/teams"""

        experiment_parse = values["task"].experiment_parse
        exp = experiment_parse.experiment_class(**experiment_parse.config)

        # HACK: Ensure that the Simulator generates output AERs for every worker
        exp.config.env_config["simulator"]["config"]["limit_extra_data"] = False

        tmp = exp.create_agents(
            platform_configs=values["teams"].platform_config,
            agent_configs=values["teams"].agent_config,
        )
        exp.config.env_config["agents"], exp.config.env_config["agent_platforms"] = tmp

        tmp = exp.create_environment()
        if hasattr(tmp.simulator, "shutdown"):
            tmp.simulator.shutdown()

        policies, train_policies, _ = exp.create_policies(tmp)

        exp.config.policy_mapping.functor = PolicyIsAgent
        exp.config.policy_mapping.config = {}

        exp.config.rllib_configs["local"]["multiagent"] = {
            "policies": policies,
            "policy_mapping_fn": exp.config.policy_mapping.functor(**exp.config.policy_mapping.config),
            "policies_to_train": train_policies,
        }

        values["rllib_config"] = exp.config.rllib_configs["local"]
        values["tune_config"] = exp.config.tune_config
        values["ray_config"] = exp.config.ray_config
        values["env_config"] = exp.config.env_config

        return values



"""
-------------------------------------------------------------------------------
The Autonomous Capabilities Team (ACT3) Deep Reinforcement Learning (D-RL) Environment

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
-------------------------------------------------------------------------------

Environment Common Module
"""


class EnvCommonValues:
    """Environment Common Values"""

    METERS_TO_FEET = 3.28084
    FEET_TO_METERS = 0.3048
    MACH_TO_METERS_PER_SECOND = 343
    NAUTICAL_MILE_TO_METERS = 1852
    NAUTICAL_MILE_TO_KILOMETERS = 1.852
    METERS_TO_NAUTICAL_MILE = 0.000539957
    DEG_TO_RAD = 0.01745329251994329576923
    RAD_TO_DEG = 57.29577951308232087721
    METERS_PER_SECOND_TO_KNOTS = 1.94384
    FEET_PER_SECOND_TO_KNOTS = 0.592484
    KNOTS_TO_METERS_PER_SECOND = 0.514444
    METERS_TO_KILOMETERS = 0.001


class EnvCommonStrings:
    """Environment Common Strings"""

    SLANT_RANGE_RATE = "range_change_rate"
    SLANT_RANGE = "range"
    ASPECT_ANGLE = "aspect_angle"
    CROSS_ANGLE = "heading_cross_angle"
    RELATIVE_ELEVATION = "relative_elevation"

    ALT_ERROR = "alt_error"
    SUM_ALT_ERROR = "sum_alt_error"
    ALT_ERROR_DIFF = "alt_error_diff"



"""
---------------------------------------------------------------------------


Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
Evaluation from Training
"""
import streamlit as st

from corl.evaluation.recording.folder import EpisodeArtifactLoggingCallbackLoader
from corl.visualization.streamlit_app.eval_while_train_layout import StreamlitPageEvaluationWhileTraining

if __name__ == "__main__":
    st.set_page_config(layout="wide")

    # #########################################################################
    # Inspects results generated by running CoRL's separate evaluation script.
    # #########################################################################

    # #########################################################################
    # Directory Selector
    if st.session_state.get("main_page_loaded"):
        st.markdown("# Explore Custom Evaluation Results")
        if not hasattr(st.session_state, "eval_while_train_page"):
            st.session_state.eval_while_train_page = StreamlitPageEvaluationWhileTraining(file_loader=EpisodeArtifactLoggingCallbackLoader)

        st.session_state.eval_while_train_page.display_directory_picker()

        if st.session_state.eval_while_train_page.full_experiment_path:
            st.session_state.eval_while_train_page.display_epoch_picker()
            if st.session_state.eval_while_train_page.epoch_selection != "":  # noqa: PLC1901
                st.session_state.eval_while_train_page.display_load_episode_artifact_option()
            if st.session_state.eval_while_train_page.tables_dict:
                st.session_state.eval_while_train_page.display_episode_metadata_section()
                st.session_state.eval_while_train_page.display_agent_dones_section()
                st.session_state.eval_while_train_page.display_platform_dones_section()
                st.session_state.eval_while_train_page.display_initial_states_section()
                st.session_state.eval_while_train_page.display_units_selector()
                st.session_state.eval_while_train_page.display_agent_steps_section()
                if st.session_state.layout_config:
                    st.session_state.eval_while_train_page.display_trajectory_animation_section(
                        **st.session_state.layout_config["streamlit_layout"]["display_trajectory_animation_section"]
                    )
                else:
                    st.warning("Missing Trajectory animation configuration. No animation is provided.")
                st.session_state.eval_while_train_page.display_space_data_table()
    else:
        st.warning("Main page must be loaded before navigating to subsequent pages. Navigate back to the first page `streamlit_app`)")



"""
---------------------------------------------------------------------------


Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
Functions for the unit options in the streamlit app
"""

import pandas as pd
import streamlit as st

from corl.libraries.units import corl_get_ureg
from corl.visualization.streamlit_app.utils import write_code_block


def unit_selection_and_table_update(tables_dict: dict[str, pd.DataFrame], obs_units_flattened: dict) -> dict:
    """Streamlit section that displays the unit selector and returns the
    tables dict with the units converted for the AgentSteps and SpaceData tables.

    Parameters
    ----------
    tables_dict : Dict
        This is expected to be EpisodeArtifactTables.tables
    obs_units_flattened : Dict[str, str]
        A dictionary that maps the attribute name to the unit name (str)

    Returns
    -------
    Dict[str, pd.DataFrame]
        The updated tables_dict
    """
    st.markdown("## Unit Conversion")
    # Check if there are any values with real units
    none_units = [unit_str == "dimensionless" for _, unit_str in obs_units_flattened.items()]
    if len(obs_units_flattened) == 0 or all(none_units):
        st.warning("No observation unit information was found or all units are dimensionless types.")
        return tables_dict

    st.markdown("**Note**: This will change both the SpaceData table and the AgentSteps table.")
    ignore_dimensions = ["dimensionless"]
    present_dimensions = []
    # attribute_name -> dimension
    attribute_name_dimension_map = {}
    obs_units_flattened = obs_units_flattened.copy()
    # Gets a list of relevant dimensions
    for attribute_name, unit_str in obs_units_flattened.items():
        if unit_str not in ignore_dimensions:
            attribute_name_dimension_map[attribute_name] = unit_str
            if unit_str not in present_dimensions:
                present_dimensions.append(unit_str)

    # Creates a dropbox selector for each (unit) dimension
    user_selected_units = {}
    columns = st.columns(4)
    col_idx = 0
    for dimension_name in present_dimensions:
        if dimension_name in ignore_dimensions:
            continue

        ureg = corl_get_ureg()
        options = []
        for u in ureg.units:
            if corl_get_ureg().get_unit(dimension_name).is_compatible_with(u) and corl_get_ureg().get_unit(u) not in options:
                options.append(str(corl_get_ureg().get_unit(u)))

        # dimension_name -> desired_units
        if col_idx % 4 == 0:
            col_idx = 0
            columns = st.columns(4)
        with columns[col_idx]:
            user_selected_units[dimension_name] = st.selectbox(label=dimension_name, options=options, key=f"selection_{dimension_name}")
        col_idx += 1

    if st.button("Convert units"):
        write_code_block("Converting units in agent steps...")
        # attribute_name -> new units
        new_units_map = {
            attribute_name: user_selected_units[dimension_name] for attribute_name, dimension_name in attribute_name_dimension_map.items()
        }

        old_to_new_units: dict[str, str] = {obs_units_flattened[k]: v for k, v in new_units_map.items() if v is not None}
        update_units_flattened_agent_steps(tables_dict=tables_dict, old_to_new_units_dict=old_to_new_units)

        write_code_block("Updating units in space table...")
        tables_dict = update_units_space_table(tables_dict=tables_dict, new_units_map=new_units_map)

    return tables_dict


def update_units_flattened_agent_steps(tables_dict: dict, old_to_new_units_dict: dict[str, str]) -> dict:
    """Converts the united values in the agents steps table.

    Parameters
    ----------
    old_to_new_units_dict: dict
        A mapping from old units to new units
    """
    dataframe = tables_dict["FlattenedAgentStep"].copy()
    if not dataframe.empty:

        def apply_unit_conversion_to_row(row):
            tmp = corl_get_ureg().Quantity(row["attribute_value"], units=row["units"])
            return tmp.m_as(old_to_new_units_dict[row["units"]])

        idx_has_units = dataframe["units"].notnull()
        dataframe.loc[idx_has_units, "attribute_value"] = dataframe.loc[idx_has_units].apply(
            lambda row: apply_unit_conversion_to_row(row) if row["units"] in old_to_new_units_dict else row["attribute_value"], axis=1
        )
        dataframe["units"] = dataframe["units"].apply(lambda x: old_to_new_units_dict.get(x, x))
        tables_dict["FlattenedAgentStep"] = dataframe
    return tables_dict


def update_units_space_table(tables_dict: dict, new_units_map: dict) -> dict:
    """Updates the units in the Space Data table

    Parameters
    ----------
    new_units_map : dict
        The key must be an attribute name and the value must be the unit to convert the value to.
    """
    dataframe = tables_dict["SpaceData"].copy()
    if not dataframe.empty:

        def apply_unit_conversion_to_row(row, prefix, column):
            tmp = corl_get_ureg().Quantity(row[f"{prefix}{column}"], units=row["_prev_units"])
            return tmp.m_as(row["units"])

        dataframe["_prev_units"] = dataframe["units"].copy()

        dataframe["units"] = dataframe.apply(lambda row: new_units_map.get(row["attribute_name"], row["units"]), axis=1)
        for column in ["min", "max"]:
            # Convert the "normalized columns" that have not been normalized when units exist
            dataframe[f"normalized_{column}"] = dataframe.apply(
                lambda row, column=column: apply_unit_conversion_to_row(row, "normalized_", column)
                # Only performs conversions when the normalized
                # and unnormalized columns match and units exist
                if row["min"] == row["normalized_min"] and row["max"] == row["normalized_max"] and row["_prev_units"]
                else row[f"normalized_{column}"],
                axis=1,
            )
            # Converts the unnormalized columns. NOTE: this must go after the normalized column.
            dataframe[column] = dataframe.apply(
                lambda row, column=column: apply_unit_conversion_to_row(row, "", column) if row["_prev_units"] else row[column], axis=1
            )

        dataframe = dataframe.drop(columns=["_prev_units"])
        tables_dict["SpaceData"] = dataframe

    return tables_dict



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
"""

from pydantic import ConfigDict

from corl.glues.base_glue import BaseAgentGlue, BaseAgentGlueValidator


class BaseWrapperGlueValidator(BaseAgentGlueValidator):
    """
    wrapped - the wrapped glue instance
    """

    wrapped: BaseAgentGlue
    model_config = ConfigDict(arbitrary_types_allowed=True)


class BaseWrapperGlue(BaseAgentGlue):
    """A base object that glues can inherit in order to "wrap" a single glue instance"""

    def __init__(self, **kwargs) -> None:
        self.config: BaseWrapperGlueValidator
        super().__init__(**kwargs)

    @staticmethod
    def get_validator() -> type[BaseWrapperGlueValidator]:
        return BaseWrapperGlueValidator

    def glue(self) -> BaseAgentGlue:
        """Get the wrapped glue instance"""
        return self.config.wrapped

    def set_agent_removed(self, agent_removed: bool = True) -> None:
        super().set_agent_removed(agent_removed)
        self.glue().set_agent_removed(agent_removed)



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
Load and agent from Checkpoint file
"""
import os
import typing
from pathlib import Path

from pydantic import BaseModel, ConfigDict, field_validator
from ray.rllib.algorithms import Algorithm
from ray.rllib.policy.policy import Policy

from corl.evaluation.loader.i_agent_loader import IAgentLoader


class PolicyCheckpoint(BaseModel, IAgentLoader):
    """Class manages loading an agent from a checkpoint file
    Because this is used/instantiated via jsonargparse, we need to use pydantic.dataclasses.dataclass here (instead of BaseModel);
    in conjunction with jsonargparse, this results in some (unfortunate) side effects:
      1. Can't directly use Path
      2. Can't call validators before some validation is performed (i.e. @validator `pre` parameter doesn't behave as expected)
        a. The primary side effect of this is that derived values must be defined as Optional, even though they will exist after validation
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)

    trained_agent_id: str | None = None
    checkpoint_filename: Path

    policy_: Policy | None = None

    @field_validator("checkpoint_filename", mode="before")
    def validate_checkpoint_filename(cls, v, values):
        """Validates the checkpoint directory"""
        assert v is not None, "checkpoint_filename must be set"
        p = Path(os.path.expandvars(str(v)))
        assert p.exists(), f"{p} does not exist"
        assert p.is_dir(), f"{p} must be a directory"

        trained_agent_id = values.data["trained_agent_id"] if values.data["trained_agent_id"] is not None else ""

        algorithm_policy_state_file = p / "policies" / trained_agent_id / "policy_state.pkl"
        policy_policy_state_file = p / "policy_state.pkl"

        if not (algorithm_policy_state_file.exists() or policy_policy_state_file.exists()):
            raise ValueError(f"No policy state found for {p}:\n\t{algorithm_policy_state_file}\n\t{policy_policy_state_file}")

        return p

    @property
    def policy(self) -> Policy:
        """
        returns the policy object given the class config

        Raises:
            RuntimeError: if the specified agent is not the checkpoint

        Returns:
            Policy: The policy from the checkpoint
        """
        if not self.policy_:
            policy = Policy.from_checkpoint(str(self.checkpoint_filename))

            if isinstance(policy, dict):
                if self.agent_id not in policy:
                    raise RuntimeError(f"agent_id '{self.agent_id}' not in '{policy}'")
                self.policy_ = policy[self.agent_id]
            elif isinstance(policy, Policy):
                self.policy_ = policy
            else:
                raise TypeError(f"Unknown type {type(policy)} for {Policy}")

        return typing.cast(Policy, self.policy_)

    @property
    def was_trained(self) -> bool:
        """
        specifies if a policy to be extract was trained in the checkpoint
        """
        return self.agent_id in self.policy.config["multiagent"]["policies_to_train"]

    @property
    def agent_id(self) -> str | None:
        if self.trained_agent_id is not None:
            return self.trained_agent_id

        if "agent_id" not in self.policy.config:
            return Path(self.checkpoint_filename).name

        return self.policy.config["agent_id"]

    @property
    def env_config(self) -> dict | None:
        return self.policy.config["env_config"]

    def apply_to_algorithm(self, algorithm: Algorithm, policy_to_apply: str) -> None:
        """Applies the weights from checkpoint file to an RLLIB algorithm

        Arguments:
            algorithm {Algorithm} -- Algorithm to apply weight file to
            policy_to_apply {str} -- Policy in the algorithm to apply weights to
        """
        if weights := self.policy.get_weights():
            self._apply_policy_to_weights(
                algorithm=algorithm,
                weights=weights,
                policy_to_apply=policy_to_apply,
            )
        else:
            raise RuntimeError(f"Policy {self.checkpoint_filename} was not trained - No weights available.")



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
"""
import abc
from collections import OrderedDict

import gymnasium
from pydantic import ConfigDict

from corl.rewards.reward_func_base import RewardFuncBase, RewardFuncBaseValidator
from corl.simulators.base_simulator_state import BaseSimulatorState


class BaseMultiWrapperRewardValidator(RewardFuncBaseValidator):
    """
    wrapped - the wrapped reward instances
    """

    wrapped: list[RewardFuncBase]
    model_config = ConfigDict(arbitrary_types_allowed=True)


class BaseMultiWrapperReward(RewardFuncBase):
    """A base object that rewards can inherit in order to "wrap" multiple reward instances"""

    def __init__(self, **kwargs) -> None:
        self.config: BaseMultiWrapperRewardValidator
        super().__init__(**kwargs)

    @staticmethod
    def get_validator() -> type[BaseMultiWrapperRewardValidator]:
        return BaseMultiWrapperRewardValidator

    def rewards(self) -> list[RewardFuncBase]:
        """Get the wrapped reward instances"""
        return self.config.wrapped

    @abc.abstractmethod
    def __call__(
        self,
        observation: OrderedDict,
        action,
        next_observation: OrderedDict,
        state: BaseSimulatorState,
        next_state: BaseSimulatorState,
        observation_space: gymnasium.Space,
        observation_units: OrderedDict,
    ) -> float:
        ...



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
"""
import typing


def cleanup(fn: typing.Callable):
    """Returns an object that will call fn when it goes out of scope"""

    class ScopedDestructor:
        """calls fn when this goes out of scope"""

        def __init__(self):
            self.abort = False
            self.clean = False

        def __del__(self):
            self.cleanup()

        def __enter__(self):
            return self

        def __exit__(self, exc_type, exc_val, exc_tb):
            self.cleanup()

        def cleanup(self):
            if not self.abort and not self.clean:
                self.clean = True
                fn()

    return ScopedDestructor()



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
Load an agent from a weight file
"""
import dataclasses
import logging

from ray.rllib.algorithms import Algorithm

from corl.evaluation.loader.i_agent_loader import IAgentLoader


@dataclasses.dataclass
class Heuristic(IAgentLoader):
    """An agent loader that makes no manipulation to a given algorithm"""

    def __post_init__(self) -> None:
        self._log: logging.Logger = logging.getLogger(Heuristic.__name__)

    def apply_to_algorithm(self, algorithm: Algorithm, policy_to_apply: str) -> None:
        """Do nothing with the algorithm given"""



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
"""
import abc
from collections import OrderedDict

import gymnasium
from pydantic import ConfigDict

from corl.rewards.reward_func_base import RewardFuncBase, RewardFuncBaseValidator
from corl.simulators.base_simulator_state import BaseSimulatorState


class BaseWrapperRewardValidator(RewardFuncBaseValidator):
    """
    wrapped - the wrapped reward instance
    """

    wrapped: RewardFuncBase
    model_config = ConfigDict(arbitrary_types_allowed=True)


class BaseWrapperReward(RewardFuncBase):
    """A base object that rewards can inherit in order to "wrap" a single reward instance"""

    def __init__(self, **kwargs) -> None:
        self.config: BaseWrapperRewardValidator
        super().__init__(**kwargs)

    @staticmethod
    def get_validator() -> type[BaseWrapperRewardValidator]:
        return BaseWrapperRewardValidator

    def reward(self) -> RewardFuncBase:
        """Get the wrapped reward instance"""
        return self.config.wrapped

    @abc.abstractmethod
    def __call__(
        self,
        observation: OrderedDict,
        action,
        next_observation: OrderedDict,
        state: BaseSimulatorState,
        next_state: BaseSimulatorState,
        observation_space: gymnasium.Space,
        observation_units: OrderedDict,
    ) -> float:
        ...



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
Classes that represent a metric generator

A metric generator is a class that generates a metric.
"""

from pydantic import BaseModel, ConfigDict, field_validator

from corl.evaluation.metrics.metric import Metric
from corl.evaluation.metrics.scopes import Scopes, from_string
from corl.evaluation.util.condition import NamedCondition


class ErrorAlert(Exception):
    """Custom exception to be raised when an alert with "error" is encountered"""


class Alert(BaseModel):
    """Representation of an active alert"""

    name: str
    type: str  # noqa: A003
    offense: Metric
    over_threshold: Metric
    model_config = ConfigDict(arbitrary_types_allowed=True)


class AlertGenerator(BaseModel):
    """Base class for a metric generator

    A metric generator is a class that is used to create a metric from parameters
    """

    name: str
    """Name of Alert
    """

    metric: str  # TODO.... that is not really a union
    """Metric to generate alert from
    """

    scope: str | Scopes
    """Scope to apply the alert to
    """

    thresholds: list[NamedCondition]
    """Thresholds to trigger specific alert types
    """

    @field_validator("scope")
    @classmethod
    def change_scope(cls, v):
        """Changes the given scope to a Scopes instance"""
        return from_string(v) if isinstance(v, str) else v

    def generate(self, metrics: dict[str, Metric]) -> list[Alert]:
        """Generate any alerts that are generated by the given thresholds

        May be multiple if there are multiple thresholds
        """
        if not isinstance(self.metric, str):
            raise RuntimeError(f'"metric" field must be string, received: {type(self.metric)}')

        # Generate the active alerts
        alerts = []
        for threshold in self.thresholds:
            if threshold.condition.func(metrics[self.metric]):
                # If an error then raise an exception
                if threshold.type == "error":
                    raise ErrorAlert(self.name)

                alerts.append(
                    Alert(
                        name=self.name,
                        type=threshold.type,
                        offense=metrics[self.metric],
                        over_threshold=metrics[self.metric] - threshold.condition.lhs,  # type: ignore[operator]
                    )
                )

        return alerts



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
"""
import logging
from collections import OrderedDict
from functools import cached_property

import numpy as np

from corl.glues.base_dict_wrapper import BaseDictWrapperGlue, BaseDictWrapperGlueValidator
from corl.glues.common.target_value import LimitConfigValidator
from corl.libraries.property import BoxProp, DictProp
from corl.libraries.units import Quantity, corl_quantity
from corl.libraries.utils import get_wrap_diff_quant


class TargetValueDifferenceValidator(BaseDictWrapperGlueValidator):
    """
    target_value: Quantity - Value to report the difference around
    value_index: - index to extract the value from an ndarray extractor
    is_wrap: bool - if the target value difference needs to wrap around the limits
    is_rad: bool - if the read observation is in radians
    method: int - TODO: fill this is
    is_abs: bool - if the difference to the target value should be reported as absolute value
    limit: LimitConfigValidator - limit configuration
    unique_name: str - optional unique name for a target value difference glue
    remove_invalid_obs: if the invalid flags should be removed from the obs
    always_valid: never output a fixed value based on a check
    that an underlying glue may have a part that is invalid
    """

    target_value: Quantity | None = None
    target_value_index: int | None = 0
    target_value_field: str | None = "direct_observation"
    value_index: int = 0
    is_wrap: bool = False
    is_rad: bool = False
    method: int = 0
    is_abs: bool = False
    limit: LimitConfigValidator
    unique_name: str = ""
    always_valid: bool = False


class TargetValueDifference(BaseDictWrapperGlue):
    """
    Glue class for an observation of some constant value
    """

    SENSOR_STR = "sensor"
    TARGET_STR = "target"

    class Fields:
        """
        field data
        """

        TARGET_VALUE_DIFF = "diff"

    def __init__(self, **kwargs) -> None:
        self.config: TargetValueDifferenceValidator
        super().__init__(**kwargs)
        keys = self.glues().keys()
        if not isinstance(self.config.target_value, Quantity) and len(self.glues()) != 2:
            raise ValueError("self.config.target_value not defined - expecting two glues to get target from 2nd glue")

        if not isinstance(self.config.target_value, Quantity) and len(self.glues()) == 2 and TargetValueDifference.TARGET_STR not in keys:
            raise KeyError(f"Expecting to see {TargetValueDifference.TARGET_STR} in keys - {keys}")

        if isinstance(self.config.target_value, Quantity) and len(self.glues()) != 1:
            raise ValueError("self.config.target_value defined - expecting one glue")

        if TargetValueDifference.SENSOR_STR not in keys:
            raise KeyError(f"Expecting to see {TargetValueDifference.SENSOR_STR} in keys - {keys}")

        self._logger = logging.getLogger(TargetValueDifference.__name__)

        if unique_name := self.config.unique_name:
            self._uname = unique_name
        else:
            wrapped_glue_name = self.glues()[TargetValueDifference.SENSOR_STR].get_unique_name()
            self._uname = None if wrapped_glue_name is None else f"{wrapped_glue_name}Diff"

    @staticmethod
    def get_validator() -> type[TargetValueDifferenceValidator]:
        """Return validator"""
        return TargetValueDifferenceValidator

    def get_unique_name(self) -> str:
        """Class method that retrieves the unique name for the glue instance"""
        return self._uname

    @cached_property
    def observation_prop(self):
        space = self.glues()[TargetValueDifference.SENSOR_STR].observation_space

        if len(space.spaces) > 1:
            raise RuntimeError("Target value difference can only wrap a glue with one output")

        new_field_name = self.Fields.TARGET_VALUE_DIFF
        for field in space.spaces:
            new_field_name = f"{field}_diff"
        return DictProp(
            spaces={
                new_field_name: BoxProp(
                    low=self.config.limit.minimum.m,
                    high=self.config.limit.maximum.m,
                    dtype=np.dtype("float32"),
                    unit=self.config.limit.unit,
                )
            }
        )

    def get_observation(self, other_obs: OrderedDict, obs_space, obs_units):
        """Get observation"""
        glue = self.glues()[TargetValueDifference.SENSOR_STR]
        obs = glue.get_observation(other_obs, obs_space, obs_units)
        target_invalid = False

        if isinstance(self.config.target_value, Quantity):
            target_value = self.config.target_value
            if np.isscalar(target_value.m):
                target_value = corl_quantity()(np.asarray([target_value.m], np.float32), self.config.target_value.u)
        else:
            # TODO make this more configurable... assumes 1 value for target
            if hasattr(self.glues()[TargetValueDifference.TARGET_STR], "_sensor"):
                target_invalid = not self.glues()[TargetValueDifference.TARGET_STR]._sensor.valid  # type: ignore # noqa: SLF001
            target_value = self.glues()[TargetValueDifference.TARGET_STR].get_observation(other_obs, obs_space, obs_units)[
                self.config.target_value_field
            ][  # type: ignore
                self.config.target_value_index
            ]

        assert isinstance(target_value, Quantity)
        target_value = target_value.to(self.config.limit.unit)

        if self.config.always_valid:
            target_invalid = False
        d = OrderedDict()

        if not isinstance(obs, dict):
            raise RuntimeError("This glue cannot handle a wrapped glue outputting something other than a dict")

        if target_invalid:
            for field in obs:
                d[f"{field}_diff"] = corl_quantity()(np.array([self.config.limit.minimum.m], dtype=np.float32), self.config.limit.unit)
        else:
            for field, value in obs.items():
                new_val = value[self.config.target_value_index] if value.m.shape else value
                diff = (
                    get_wrap_diff_quant(target_value, new_val, self.config.method)
                    if self.config.is_wrap
                    else corl_quantity()(np.asarray(target_value.m - new_val.m), target_value.u)
                )
                unit_name = self.observation_prop.spaces[f"{field}_diff"].get_units()
                diff = diff.to(unit_name)

                if self.config.is_abs:
                    diff = abs(diff.m)

                self._logger.debug(f"{TargetValueDifference.__name__}::obs = {value}")
                self._logger.debug(f"{TargetValueDifference.__name__}::target = {target_value}")
                self._logger.debug(f"{TargetValueDifference.__name__}::diff = {diff}")

                d[f"{field}_diff"] = corl_quantity()(np.asarray(diff.m.astype(np.float32)), unit_name)

        return d



"""
This module implements the Reward Functions and Reward Validators specific to the 1D Docking task.
"""

from collections import OrderedDict

import gymnasium
import numpy as np

from corl.libraries.units import Quantity
from corl.rewards.reward_func_base import RewardFuncBase, RewardFuncBaseValidator
from corl.simulators.base_simulator_state import BaseSimulatorState
from corl.simulators.common_platform_utils import get_platform_by_name, get_sensor_by_name


class DockingRewardValidator(RewardFuncBaseValidator):
    """
    This Validator ensures the DockingReward's config defines values relevant to successful and unsuccessful docking
    attempts.
    """

    success_reward: float
    timeout_reward: float
    distance_reward: float
    crash_reward: float
    timeout: float
    docking_region_radius: Quantity
    max_goal_distance: Quantity
    velocity_threshold: Quantity
    position_sensor_name: str
    velocity_sensor_name: str


class DockingReward(RewardFuncBase):
    """
    This Reward Function is responsible for calculating the reward (or penalty) associated with a given docking attempt.
    """

    REQUIRED_UNITS = {"docking_region_radius": "meter", "velocity_threshold": "meter / second", "max_goal_distance": "meter"}

    def __init__(self, **kwargs) -> None:
        self.config: DockingRewardValidator
        super().__init__(**kwargs)

    @staticmethod
    def get_validator() -> type[DockingRewardValidator]:
        """
        Method to return class's Validator.
        """
        return DockingRewardValidator

    def __call__(
        self,
        observation: OrderedDict,
        action,
        next_observation: OrderedDict,
        state: BaseSimulatorState,
        next_state: BaseSimulatorState,
        observation_space: gymnasium.Space,
        observation_units: OrderedDict,
    ) -> float:
        """
        This method determines if the agent has succeeded or failed and returns an appropriate reward.

        Parameters
        ----------
        observation : OrderedDict
            The observations available to the agent from the previous state.
        action
            The last action performed by the agent.
        next_observation : OrderedDict
            The observations available to the agent from the current state.
        state : StateDict
            The previous state of the simulation.
        next_state : StateDict
            The current state of the simulation.
        observation_space : StateDict
            The agent's observation space.
        observation_units : StateDict
            The units corresponding to values in the observation_space

        Returns
        -------
        reward
            The agent's reward for their docking attempt.
        """
        value = 0.0

        deputy = get_platform_by_name(next_state, self.config.platform_names[0])

        position_sensor = get_sensor_by_name(deputy, self.config.position_sensor_name)  # type: ignore
        velocity_sensor = get_sensor_by_name(deputy, self.config.velocity_sensor_name)  # type: ignore

        position = position_sensor.get_measurement().m
        velocity = velocity_sensor.get_measurement().m
        sim_time = deputy.sim_time  # type: ignore

        chief_position = np.array([0])
        docking_region_radius = self.config.docking_region_radius.m

        distance = abs(position - chief_position)
        in_docking = distance <= docking_region_radius

        max_velocity_exceeded = self.config.velocity_threshold.m < velocity

        if sim_time > self.config.timeout:
            # episode reached max time
            value = self.config.timeout_reward
        elif distance >= self.config.max_goal_distance.m:
            # agent exceeded max distance from goal
            value = self.config.distance_reward
        elif in_docking and max_velocity_exceeded:
            # agent exceeded velocity constraint within docking region
            value = self.config.crash_reward
        elif in_docking and not max_velocity_exceeded:
            # agent safely made it to the docking region
            value = self.config.success_reward
            if self.config.timeout:
                # Add time reward component, if timeout specified
                value += 1 - (sim_time / self.config.timeout)

        return value



"""
This module extends corl.simulators.base_platform.BasePlatform to create a simple one dimensional
docking platform.
"""


import numpy as np

from corl.libraries.units import corl_get_ureg
from corl.simulators.base_platform import BasePlatform, BasePlatformValidator
from corl.simulators.docking_1d.entities import Deputy1D


class Docking1dPlatformValidator(BasePlatformValidator):
    """Docking1dPlatformValidator

    Parameters
    ----------
        platform: Deputy1D
            Deputy associated with the CoRL Docking1dPlatform
    """

    platform: Deputy1D


class Docking1dPlatform(BasePlatform):  # noqa: PLW1641
    """
    A platform representing a spacecraft operating under Double Integrator dynamics.
    Allows for saving an action to the platform for when the platform needs
    to give an action to the environment during the environment step function

    Parameters
    ----------
    platform_name : str
        Name of the platform
    platform : sim_entity
        Backend simulation entity associated with the platform
    platform_config : dict
        Platform-specific configuration dictionary
    """

    def __init__(self, **kwargs) -> None:
        self.config: Docking1dPlatformValidator
        super().__init__(**kwargs)

        self._platform = self.config.platform
        self._last_applied_action = corl_get_ureg().Quantity(np.array([0], dtype=np.float32), "newtons")  # thrust
        self._sim_time = 0.0

    @staticmethod
    def get_validator() -> type[Docking1dPlatformValidator]:
        return Docking1dPlatformValidator

    def __eq__(self, other):
        if isinstance(other, Docking1dPlatform):
            eq = (self.velocity == other.velocity).all()
            eq = eq and (self.position == other.position).all()
            return eq and self.sim_time == other.sim_time
        return False

    def get_applied_action(self):
        """
        returns the action stored in this platform

        Returns:
            typing.Any -- any sort of stored action
        """
        return self._last_applied_action

    def save_action_to_platform(self, action):
        """
        saves an action to the platform if it matches
        the action space

        Arguments:
            action typing.Any -- The action to store in the platform
        """
        if isinstance(action.m, np.ndarray) and len(action.m) == 1:
            self._last_applied_action = action

    @property
    def position(self):
        """
        The position of the platform

        Returns
        -------
        np.ndarray
            The position vector of the platform
        """
        return self._platform.position

    @property
    def velocity(self):
        """
        The velocity of the platform

        Returns
        -------
        np.ndarray
            The velocity vector of the platform
        """
        return self._platform.velocity

    @property
    def sim_time(self):
        """
        The current simulation time in seconds.

        Returns
        -------
        float
            Current simulation time
        """
        return self._sim_time

    @sim_time.setter
    def sim_time(self, time):
        self._sim_time = time

    @property
    def operable(self):
        return True



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
"""

from collections import OrderedDict

import gymnasium


def gymnasium_space_sort(space: gymnasium.spaces.Space):
    """
    space:
        Any gymnasium.spaces.Space (or nested hierarchy of spaces) to be sorted. This sorts
        any gymnasium.space.Dict in the hierarchy in KEY order.
        This is necessary as a workaround to a bug in ray/rllib which causes the action space that
        is generated from a dictionary to be
        in KEY order (the result of this line:
        https://github.com/ray-project/ray/blob/9c26d6c6de3016eb7409db8eca51f8cac4b2944e/rllib/utils/exploration/random.py#L119
        which uses https://tree.readthedocs.io/en/latest/api.html#tree.map_structure, which internally uses
        https://tree.readthedocs.io/en/latest/api.html#tree.flatten which, per the docs, causes the returned map to be in KEY order).
    """
    if isinstance(space, gymnasium.spaces.Dict):
        space_dict: gymnasium.spaces.Dict = space
        for key in space_dict.spaces:
            sub_space = space_dict.spaces[key]
            gymnasium_space_sort(sub_space)
        space.spaces = OrderedDict(sorted(space_dict.spaces.items()))
    if isinstance(space, gymnasium.spaces.Tuple):
        space_tuple: gymnasium.spaces.Tuple = space
        for sub_space in space_tuple.spaces:
            gymnasium_space_sort(sub_space)



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
"""


from pydantic import BaseModel, Extra


class BaseSimulatorState(BaseModel, extra=Extra.forbid):
    """
    Pydantic model containing a BaseClass for the state a simulator should return
    Other simulators may subclass this to return custom information
    """

    sim_platforms: dict
    sim_time: float
    user_data: dict = {}
    # episode_history: defaultdict[str, deque] = defaultdict(partial(deque, maxlen=1000))
    episode_state: dict = {}
    agent_episode_state: dict = {}
    simulator_info: dict = {}
    sim_update_rate_hz: float = 1



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
Scripted Action Policy
"""
import typing

import flatten_dict
import numpy as np
from pydantic import ConfigDict, field_validator, model_validator

from corl.libraries.env_space_util import EnvSpaceUtil
from corl.policies.custom_policy import CustomPolicy, CustomPolicyValidator


class ScriptedActionPolicyValidator(CustomPolicyValidator):
    """Validator for the ScriptedActionPolicy"""

    control_times: list[float]
    control_values: list[dict]

    missing_action_policy: typing.Literal["default_action", "repeat_last_action"]
    default_action: dict | None = None
    model_config = ConfigDict(arbitrary_types_allowed=True)

    @field_validator("control_times")
    @classmethod
    def sort_control_times(cls, v):
        """Ensures that control_times are in order"""
        assert v == sorted(v), "control_times must be in order"
        return v

    @staticmethod
    def convert_control_value(controls, controller_key_paths, sample_control):
        """converts the controls into a dict"""

        flat_sample_control = flatten_dict.flatten(sample_control)

        assert len(controls) == len(controller_key_paths), "mismatch between number of controllers and length of control values"
        assert len(controller_key_paths) == len(flat_sample_control), "mismatch between number of controllers and the action_space"

        flat_control_dict = {}
        for i, ctrl_value in enumerate(controls):
            try:
                # The flattened sample control key is expected to be a tuple
                # of the flattened keys.
                controller_key = tuple(controller_key_paths[i]) if isinstance(controller_key_paths[i], list) else controller_key_paths[i]

                sample_value = flat_sample_control[controller_key]
                if isinstance(sample_value, np.ndarray):
                    flat_control_dict[controller_key] = np.add(sample_value * 0, ctrl_value, dtype=sample_value.dtype)
                else:
                    flat_control_dict[controller_key] = type(sample_value)(ctrl_value)
            except Exception as e:  # noqa: BLE001,PERF203
                raise RuntimeError(f"@idx: {i}, controller: {controller_key}, control_value: {ctrl_value}") from e
        return flatten_dict.unflatten(flat_control_dict)

    @staticmethod
    def convert_control_values(action_space, controller_key_paths, control_times, controls_list) -> list[dict]:
        """converts the input_control values into dictionary and validates it against the action space"""
        assert len(controls_list) == len(control_times), "mismatch between number of control_times and control_values"

        sample_control = action_space.sample()

        converted_control_list: list[dict] = []
        for controls in controls_list:
            control_dict = ScriptedActionPolicyValidator.convert_control_value(controls, controller_key_paths, sample_control)

            EnvSpaceUtil.deep_sanity_check_space_sample(action_space, control_dict)

            converted_control_list.append(control_dict)

        return converted_control_list

    @model_validator(mode="before")
    def validate_control_and_action_values(cls, values):
        assert isinstance(values, dict)
        assert "controllers" in values and "control_times" in values

        values["control_values"] = ScriptedActionPolicyValidator.convert_control_values(
            values["act_space"], values["controllers"], values["control_times"], values["controls_list"]
        )

        """validates that the default_action is consistent with the missing_action_policy"""
        default_action = values.get("default_action", None)
        missing_action_policy = values["missing_action_policy"]
        if missing_action_policy == "default_action":
            assert default_action is not None, "default_action is required when using the default_action missing_action_policy"

            action_space = values["act_space"]
            values["default_action"] = ScriptedActionPolicyValidator.convert_control_value(
                default_action, values["controllers"], action_space.sample()
            )
        elif missing_action_policy == "repeat_last_action":
            assert values["control_times"][0] == 0, f"missing control_time for t={values['reset_time']}"
        else:
            assert default_action is None, "default_action is invalid except when using the default_action missing_action_policy"
        return values


class ScriptedActionPolicy(CustomPolicy):
    """Scripted action policy."""

    def __init__(self, observation_space, action_space, config) -> None:
        super().__init__(observation_space, action_space, config)

        self._input_index: int
        self._last_action: dict

    @staticmethod
    def get_validator() -> type[ScriptedActionPolicyValidator]:
        """
        Get the validator for this experiment class,
        the kwargs sent to the experiment class will
        be validated using this object and add a self.config
        attr to the experiment class
        """
        return ScriptedActionPolicyValidator

    def _reset(self):
        super()._reset()
        self._input_index = 0

        if self.validated_config.default_action is not None:
            self._last_action = self.validated_config.default_action
        else:
            self._last_action = EnvSpaceUtil.get_mean_sample_from_space(self.validated_config.act_space)

    def custom_compute_actions(
        self,
        obs,
        platform_obs,
        state=None,
        prev_action=None,
        prev_reward=None,
        info=None,
        explore=None,
        timestep=None,
        sim_time=None,
        agent_id=None,
        epp_info=None,
        episode=None,
        **kwargs,
    ):
        for control_index in range(self._input_index, len(self.validated_config.control_times)):
            control_time = self.validated_config.control_times[control_index]
            if sim_time >= control_time:
                # apply control_list to controls
                control_values = self.validated_config.control_values[control_index]

                self._input_index = control_index + 1
                self._last_action = control_values
                return control_values

            break

        if self.validated_config.missing_action_policy == "repeat_last_action":
            return self._last_action

        self._last_action = self.validated_config.default_action
        return self.validated_config.default_action



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
"""

import copy
import typing
from collections import OrderedDict

import gymnasium.utils.seeding
from pydantic import BaseModel
from ray.rllib.env.multi_agent_env import MultiAgentEnv

from corl.agents.base_agent import BaseAgent
from corl.episode_parameter_providers import EpisodeParameterProvider
from corl.libraries.env_space_util import EnvSpaceUtil
from corl.libraries.units import UnitRegistryConfiguration, corl_set_ureg
from corl.simulators.base_simulator import BaseSimulator, BaseSimulatorState


class BaseCorlMultiAgentEnvValidator(BaseModel):
    """Validation model for the inputs of BaseCorlMultiAgentEnv"""

    disable_action_history: bool = False  # disables the saving of actions during episode
    unit_reg_config: UnitRegistryConfiguration = UnitRegistryConfiguration()


class BaseCorlMultiAgentEnv(MultiAgentEnv):
    """Abstract base CoRL environment"""

    episode_parameter_provider_name: str = "environment"

    def __init__(self, config: dict):
        """Initialize base env"""
        env_config = copy.deepcopy(config)
        try:
            env_config_vars = vars(env_config)
        except TypeError:
            env_config_vars = {}

        ureg_validator = BaseCorlMultiAgentEnvValidator(unit_reg_config=env_config.get("unit_reg_config", UnitRegistryConfiguration()))
        ureg = ureg_validator.unit_reg_config.create_registry_from_args()
        corl_set_ureg(ureg)

        self.config: BaseCorlMultiAgentEnvValidator = self.get_validator()(**env_config, **env_config_vars)

        self._info: typing.OrderedDict = OrderedDict()
        self._done_info: typing.OrderedDict = OrderedDict()
        self._reward_info: typing.OrderedDict = OrderedDict()
        self._actions: list[typing.Any] = []
        self._observation: typing.OrderedDict = OrderedDict()
        self._agent_dict: dict[str, BaseAgent] = {}
        self._state: BaseSimulatorState
        self._simulator: BaseSimulator
        self._episode_id: int | None = None

    @staticmethod
    def get_validator() -> type[BaseCorlMultiAgentEnvValidator]:
        """Get the validator for this class."""
        return BaseCorlMultiAgentEnvValidator

    def post_process_trajectory(self, agent_id, batch, episode, policy):
        """easy accessor for calling post process trajectory
            correctly

        Arguments:
            agent_id: agent id
            batch: post processed Batch - be careful modifying
        """
        raise NotImplementedError

    @property
    def info(self) -> OrderedDict:
        """[summary]

        Returns:
            Union[OrderedDict, None] -- [description]
        """
        return self._info

    @property
    def done_info(self) -> OrderedDict:
        """[summary]

        Returns
        -------
        Union[OrderedDict, None]
            [description]
        """
        return self._done_info

    @property
    def reward_info(self) -> OrderedDict:
        """[summary]

        Returns
        -------
        Union[OrderedDict, None]
            [description]
        """
        return self._reward_info

    @property
    def state(self) -> BaseSimulatorState:
        """
        state of platform object.  Current state.

        Returns
        -------
        BaseSimulatorState
            the dict storing the current state of environment.
        """
        return self._state

    @property
    def simulator(self) -> BaseSimulator:
        """
        simulator simulator instance

        Returns
        -------
        BaseSimulator
            The simulator instance in the base
        """
        return self._simulator

    @property
    def observation(self) -> OrderedDict:
        """
        observation get the observation for the agents in this environment

        Returns
        -------
        OrderedDict
            the dict holding the observations for the agents
        """
        return self._observation

    @property
    def actions(self) -> list[typing.Any]:
        """
        get the current episode parameter provider episode id

        Returns
        -------
        list
            actions
        """
        return self._actions

    @property
    def episode_id(self) -> int | None:
        """
        get the current episode parameter provider episode id

        Returns
        -------
        int or None
            the episode id
        """
        return self.episode_id

    @property
    def local_variable_store(self) -> dict[str, typing.Any]:
        """
        get the local variable store

        Returns
        -------
        Dict
            the local variable store episode id
        """
        raise NotImplementedError

    @property
    def agent_dict(self) -> dict[str, BaseAgent]:
        """
        get the agent dict

        Returns
        -------
        Dict
            agents
        """
        return self._agent_dict

    @property
    def epp(self) -> EpisodeParameterProvider:
        """
        get the Episode Parameter Provider

        Returns
        -------
        EpisodeParameterProvider
        """
        raise NotImplementedError

    @property
    def glue_info(self) -> OrderedDict:
        """[summary]

        Returns:
            Union[OrderedDict, None] -- [description]
        """
        return self._info

    @staticmethod
    def _sanity_check(space: gymnasium.Space, space_sample: EnvSpaceUtil.sample_type) -> None:
        """
        Sanity checks a space_sample against a space
        1. Check to ensure that the sample from the integration base
           Fall within the expected range of values.

        Note: space_sample and space expected to match up on
        Key level entries

        Parameters
        ----------
        space: gymnasium.spaces.Space
            the space to check the sample against
        space_sample: EnvSpaceUtil.sample_type
            the sample to check if it is actually in the bounds of the space

        Returns
        -------
        OrderedDict:
            the scaled_observations
        """
        if not space.contains(space_sample):
            EnvSpaceUtil.deep_sanity_check_space_sample(space, space_sample)



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
Extra callbacks that can be for evaluating during training to
generate episode artifacts.
"""
from corl.evaluation import serialize_platforms
from corl.evaluation.runners.section_factories.engine.rllib.episode_artifact_logging_callback import EpisodeArtifactLoggingCallback


class PongEpisodeArtifactLogger(EpisodeArtifactLoggingCallback):
    """Logs episode artifacts for each evaluation epoch during training"""

    def platform_serializer(self):  # noqa: PLR6301
        return serialize_platforms.Serialize_Pong()


class Docking1dEpisodeArtifactLogger(EpisodeArtifactLoggingCallback):
    """Logs episode artifacts for each evaluation epoch during training"""

    def platform_serializer(self):  # noqa: PLR6301
        return serialize_platforms.serialize_Docking_1d()



""" non grouped utils
"""
import math
import typing
from string import digits

from corl.libraries.units import Quantity, corl_quantity


# TODO FIX TTHIS SHOULD BE A SIMPLE WRAP ON A QUANTITY
def get_wrap_diff_quant(angle0: Quantity, angle1: Quantity, method: int) -> Quantity:
    diff = angle0.to("degree").m - angle1.to("deg").m
    if method == 1:
        # Compute the diff as abs min angle (always positive)
        abs_diff_mod_360 = abs(diff) % 360
        result = 360 - abs_diff_mod_360 if abs_diff_mod_360 > 180 else abs_diff_mod_360
    else:
        # Compute the diff as min angle maintain sign for direction
        result = (diff + 180) % 360 - 180

    return corl_quantity()(result, "degree")


def get_wrap_diff(A: float, B: float, is_rad: bool, method: int) -> float:
    """Returns the min diff angle

        RAD A	         Deg A   RAD A	       Deg B   abs_diff_mod_360	Diff
        1.047197551	      60     2.094395102    120	    60	             60
        2.094395102	     120 	 1.047197551     60	    60	             60
        -2.094395102	-120	-1.047197551    -60	    60	             60
        6.108652382	     350	 0.1745329252    10	   340	             20
        -2.967059728	-170	 2.967059728    170	   340	             20

    Arguments:
        A {float} -- Angle 1 - rad or deg
        B {float} -- Angle 2 - rad or deg

    Returns:
        float -- the min diff angle
    """

    # Convert to degrees if needed.
    temp_a = math.degrees(A) if is_rad else A
    temp_b = math.degrees(B) if is_rad else B

    if method == 1:
        # Compute the diff as abs min angle (always positive)
        abs_diff_mod_360 = abs(temp_a - temp_b) % 360
        result = 360 - abs_diff_mod_360 if abs_diff_mod_360 > 180 else abs_diff_mod_360
    else:
        # Compute the diff as min angle maintain sign for direction
        diff = temp_a - temp_b
        result = (diff + 180) % 360 - 180

    # Return the diff in deg if radians if needed.
    return math.radians(result) if is_rad else result


Str_Tuple_Str = typing.TypeVar("Str_Tuple_Str", str, tuple[str, ...], str | tuple[str, ...])


def replace_magic_strings(s: Str_Tuple_Str, *, platform_name: str | None = None) -> Str_Tuple_Str:
    """Replace magic strings with their value.

    Necessary data is provided as optional keyword arguments.  Therefore, clients can determine
    which keywords they support by which pieces of information they provide.

    Known keywords:
      - %%PLATFORM%% - The name of the current platform
      - %%SIDE%% - The name of the current side, defined as the platform name with digits removed.

    Parameters
    ----------
    s : str, Tuple[str]
        String or tuple of strings in which to make replacements
    platform_name : Optional[str]
        If provided, there is support for the %%PLATFORM%% and %%SIDE%% arguments.
    """

    if platform_name is not None:
        remove_digits = str.maketrans("", "", digits)
        side = platform_name.translate(remove_digits)

        def replace_platform_side(ss: str, platform_name: str, side: str) -> str:
            return ss.replace("%%SIDE%%", side).replace("%%PLATFORM%%", platform_name)

        if isinstance(s, str):
            s = replace_platform_side(s, platform_name, side)
        else:
            s = tuple(replace_platform_side(x, platform_name, side) for x in s)

    return s



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
"""

import contextlib
import copy
import fractions
import json
import logging
import math
import os
import pickle
import sys
import typing
import warnings
from collections import OrderedDict, defaultdict
from functools import partial
from graphlib import TopologicalSorter
from itertools import chain
from json import JSONEncoder, dumps, loads
from pathlib import Path
from typing import Annotated

import deepmerge
import flatten_dict
import gymnasium.spaces
import gymnasium.utils.seeding
import numpy as np
import tree
from pydantic import (
    AfterValidator,
    BaseModel,
    BeforeValidator,
    ConfigDict,
    Field,
    NonNegativeInt,
    PositiveInt,
    StringConstraints,
    TypeAdapter,
    field_validator,
)

from corl.agents.base_agent import AgentParseInfo
from corl.dones.done_func_base import DoneFuncBase, SharedDoneFuncBase
from corl.dones.episode_length_done import EpisodeLengthDone
from corl.environment.base_multi_agent_env import BaseCorlMultiAgentEnv, BaseCorlMultiAgentEnvValidator
from corl.environment.utils import env_creation
from corl.environment.utils.obs_buffer import ObsBuffer
from corl.environment.utils.space_sort import gymnasium_space_sort
from corl.episode_parameter_providers import EpisodeParameterProvider
from corl.libraries.collection_utils import get_dictionary_subset
from corl.libraries.env_space_util import EnvSpaceUtil
from corl.libraries.environment_dict import DoneDict
from corl.libraries.factory import Factory
from corl.libraries.functor import Functor, ObjectStoreElem
from corl.libraries.parameters import ConstantParameter, Parameter, ParameterWrapper
from corl.libraries.plugin_library import PluginLibrary
from corl.libraries.property import DictProp
from corl.libraries.units import Quantity, corl_quantity
from corl.libraries.utils import replace_magic_strings
from corl.simulators.base_platform import BasePlatform
from corl.simulators.base_simulator import BaseSimulator, BaseSimulatorState, validation_helper_units_and_parameters

warnings.filterwarnings("ignore", category=np.VisibleDeprecationWarning)
warnings.filterwarnings("ignore", category=RuntimeWarning)


def check_done(v) -> None:
    """Check if dones subclass DoneFuncBase"""
    if not issubclass(v.functor, DoneFuncBase):
        raise TypeError(f"Done functors must subclass DoneFuncBase, but done {v.name} is of type {v.functor}")
    if issubclass(v.functor, EpisodeLengthDone):
        raise ValueError("Cannot specify EpisodeLengthDone as it is automatically added")
    return v


def check_shared(v):
    """Check if dones subclass SharedDoneFuncBase"""
    if not issubclass(v.functor, SharedDoneFuncBase):
        raise TypeError(f"Shared Done functors must subclass SharedDoneFuncBase, but done {v.name} is of type {v.functor}")
    return v


class EnvironmentDoneValidator(BaseModel):
    """Validation model for the dones of ACT3MultiAgentEnv"""

    world: list[Annotated[Functor, AfterValidator(check_done)]] = []
    task: dict[str, list[Annotated[Functor, AfterValidator(check_done)]]] = {}
    shared: list[Annotated[Functor, AfterValidator(check_shared)]] = []


class ACT3MultiAgentEnvEppParameters(BaseModel):
    """
    world: typing.Dict[str, typing.Dict[str, typing.Any]] = {}
        keys: done name, parameter name

    task: typing.Dict[str, typing.Dict[str, typing.Dict[str, typing.Any]]] = {}
        keys: agent name, done name, parameter name

    shared: typing.Dict[str, typing.Dict[str, typing.Any]] = {}
        keys: done name, parameter name

    reference_store: typing.Dict[str, typing.Any] = {}
        keys: reference name

    simulator_reset: typing.Dict[str, typing.Any] = {}
        keys: whatever the simulator wants, but it needs to be kwargs to simulator reset
    """

    world: dict[str, dict[str, typing.Any]] = {}
    task: dict[str, dict[str, dict[str, typing.Any]]] = {}
    shared: dict[str, dict[str, typing.Any]] = {}
    reference_store: dict[str, typing.Any] = {}
    simulator_reset: dict[str, typing.Any] = {}

    @staticmethod
    def _validate_leaves_are_parameters(obj):
        if isinstance(obj, dict):
            for value in obj.values():
                ACT3MultiAgentEnvEppParameters._validate_leaves_are_parameters(value)
        elif not isinstance(obj, Parameter | ParameterWrapper):
            raise TypeError(f"Invalid type: {type(obj)} (required type: {Parameter.__qualname__})")

    @field_validator("world", "task", "shared", "reference_store", "simulator_reset")
    @classmethod
    def validate_leaves_are_parameters(cls, v):
        """
        checks to make sure outer most leaf nodes of config are parameters
        """
        ACT3MultiAgentEnvEppParameters._validate_leaves_are_parameters(v)
        return v


def resolve_reference_store_factory(v):
    """
    allow the user to short hand constant parameter to {"value": X, "units": Y}
    """
    if isinstance(v, dict) and "value" in v and ("units" in v or "unit" in v):
        v = {"type": ConstantParameter, "config": {"value": v["value"], "units": v.get("units", v.get("unit"))}}
    return Factory.resolve_factory(v, {})


class ACT3MultiAgentEnvValidator(BaseCorlMultiAgentEnvValidator):
    """Validation model for the inputs of ACT3MultiAgentEnv"""

    num_workers: NonNegativeInt = 0
    worker_index: NonNegativeInt = 0
    vector_index: NonNegativeInt | None = None
    remote: bool = False
    deep_sanity_check: bool = True
    save_state_pickle: bool = True

    horizon: PositiveInt = 1000
    sanity_check_obs: PositiveInt = 50
    sensors_grid: list | None = None
    plugin_paths: list[str] = []

    # Regex allows letters, numbers, underscore, dash, dot
    # Regex in output_path validator also allows forward slash
    # Empty string is not allowed
    TrialName: Annotated[str, StringConstraints(pattern="^[\\w\\.-]+$")] | None = None
    output_date_string: Annotated[str, StringConstraints(pattern="^[\\w\\.-]+$")] | None = None
    skip_pbs_date_update: bool = False
    # MyPy error ignored because it is handled by the pre-validator
    output_path: str = None  # type: ignore[assignment]

    agent_platforms: dict
    agents: dict[str, AgentParseInfo]

    simulator: Factory
    other_platforms: dict[str, dict[str, typing.Any]] = {}

    reference_store: dict[str, Annotated[ObjectStoreElem, BeforeValidator(resolve_reference_store_factory)]] = {}
    dones: EnvironmentDoneValidator = Field(default=EnvironmentDoneValidator(), validate_default=True)
    end_episode_on_first_agent_done: bool = False
    simulator_reset_parameters: dict[str, typing.Any] = {}

    episode_parameter_provider: Factory
    episode_parameter_provider_parameters: ACT3MultiAgentEnvEppParameters = Field(default=None, validate_default=True)  # type: ignore
    epp_registry: dict[str, EpisodeParameterProvider] = Field(default=None, validate_default=True)  # type: ignore

    max_agent_rate: int = 20  # the maximum rate (in Hz) that an agent may be run at
    timestep_epsilon: float = 1e-3
    sim_warmup_seconds: int = 0  # number of seconds to step simulator before getting initial obs

    # dict containing logging data useful to the external system, this will be sent to simulator::mark_episode_done
    metadata: dict = {}

    @property
    def epp(self) -> EpisodeParameterProvider:
        """
        return the current episode parameter provider
        """
        return self.epp_registry[ACT3MultiAgentEnv.episode_parameter_provider_name]

    model_config = ConfigDict(arbitrary_types_allowed=True)

    @field_validator("plugin_paths")
    @classmethod
    def add_plugin_paths(cls, v):
        """Use the plugin path attribute to initialize the plugin library."""
        # @calong NOTES ON THIS CODE: DO NOT REMOVE the reraise
        # something about this code is strange when exceptions are raised and
        # either pydantic is eating the exception if something is not reraised (most likely)
        # or there is some dark magic causing exceptions to not happen, this code is a
        # simple way to make sure the exception is always raised and bypasses whatever
        # is going on, otherwise if an exception is raised the plugin process stops silently
        # and the env will fail to find a plugin and you will be left very confused
        try:
            PluginLibrary.add_paths(v)
        except Exception as e:  # noqa: BLE001
            raise RuntimeError("A plugin along your plugin paths failed to load due to the following error") from e
        return v

    @field_validator("output_path")
    def create_output_path(cls, v, values):
        """Build the output path."""

        v = v or "/tmp/data/act3/ray_results"  # noqa: S108
        adapter = TypeAdapter(Annotated[str, StringConstraints(pattern=r"^[\w/\.-]+$")])
        v = adapter.validate_python(v)

        if values.data["TrialName"] is not None:
            trial_prefix = (
                "" if values.data["skip_pbs_date_update"] else os.environ.get("PBS_JOBID", os.environ.get("TRIAL_NAME_PREFIX", ""))
            )

            v = os.path.join(v, f"{trial_prefix}-{values.data['TrialName']}") if trial_prefix else os.path.join(v, values.data["TrialName"])

        if values.data["output_date_string"] is not None and not values.data["skip_pbs_date_update"]:
            v = os.path.join(v, values.data["output_date_string"])

        v = os.path.abspath(v)
        v = os.path.join(v, str(values.data["worker_index"]).zfill(4))
        if values.data["vector_index"] is not None:
            v = os.path.join(v, str(values.data["vector_index"]).zfill(4))
        os.makedirs(v, exist_ok=True)

        return v

    @field_validator("simulator", mode="before")
    @classmethod
    def resolve_simulator_plugin(cls, v):
        """Determine the simulator from the plugin library."""
        assert isinstance(v, dict)
        try:
            v["type"]
        except (TypeError, KeyError):
            # Let pydantic print out an error when there is no type field
            return v

        match = PluginLibrary.FindMatch(v["type"], {})
        if not issubclass(match, BaseSimulator):
            raise TypeError(f"Simulator must subclass BaseSimulator, but is is of type {v['type']}")

        return {"type": match, "config": v.get("config")}

    @field_validator("agents")
    def agents_not_empty(cls, v, values):
        """Ensure that at least one agent exists"""
        if len(v) == 0:
            raise RuntimeError("No agents exist")

        agent_platforms = set(values.data["agent_platforms"])
        for agent_name, agent in v.items():
            if not all(platform_name in agent_platforms for platform_name in agent.platform_names):
                raise RuntimeError(
                    f"Agent {agent_name} is looking for platforms {agent.platform_names} "
                    f"but the only platforms available are {agent_platforms}"
                )

            if any(agent_name == platform_name for platform_name in agent.platform_names):
                raise RuntimeError(
                    f"agent '{agent_name}' has the same name as one of it's platforms {agent.platform_names}.\n"
                    " This leads to easily misconfigured code and is therefore not allowed."
                )
        return v

    @field_validator("dones")
    def agents_match(cls, v, values):
        """Ensure that platform in task dones match provided platforms"""
        # No extra agents in task dones
        for platform in v.task:
            if platform not in values.data["agent_platforms"]:
                raise RuntimeError(f"Platform {platform} lists a done condition but is not an allowed platform")

        # Task dones exist for all agents.  Make empty ones if necessary
        for platform in values.data["agent_platforms"]:
            if platform not in v.task:
                v.task[platform] = {}

        return v

    @field_validator("simulator_reset_parameters", mode="before")
    @classmethod
    def update_units_and_parameters(cls, v):
        """Update simulation reset parameters to meet base simulator requirements."""
        return validation_helper_units_and_parameters(v)

    @field_validator("episode_parameter_provider_parameters", mode="before")
    def build_episode_parameter_provider_parameters(cls, _v, values) -> ACT3MultiAgentEnvEppParameters:
        """Create the episode parameter provider for this configuration"""

        for key in ["reference_store", "dones", "simulator_reset_parameters"]:
            assert key in values.data

        reference_parameters: dict[str, Parameter | ParameterWrapper] = {
            ref_name: ref_value
            for ref_name, ref_value in values.data["reference_store"].items()
            if isinstance(ref_value, Parameter | ParameterWrapper)
        }
        world_parameters: dict[str, dict[str, Parameter | ParameterWrapper]] = {}
        for functor in values.data["dones"].world:
            functor.add_to_parameter_store(world_parameters)

        task_parameters: dict[str, dict[str, dict[str, Parameter | ParameterWrapper]]] = {}
        for agent, task_dones in values.data["dones"].task.items():
            agent_task_parameters: dict[str, dict[str, Parameter | ParameterWrapper]] = {}
            for functor in task_dones:
                functor.add_to_parameter_store(agent_task_parameters)
            task_parameters[agent] = agent_task_parameters

        shared_parameters: dict[str, dict[str, Parameter | ParameterWrapper]] = {}
        for functor in values.data["dones"].shared:
            functor.add_to_parameter_store(shared_parameters)

        sim_parameters_flat = {
            name: param
            for name, param in flatten_dict.flatten(values.data["simulator_reset_parameters"]).items()
            if isinstance(param, Parameter | ParameterWrapper)
        }
        for param_name, param_value in sim_parameters_flat.items():
            if param_name[0] != "platforms":
                continue

            for key, dep_param in param_value.config.dependent_parameters.items():
                param_value.config.dependent_parameters[key] = replace_magic_strings(dep_param, platform_name=param_name[1])

        sim_parameters = flatten_dict.unflatten(sim_parameters_flat)

        return ACT3MultiAgentEnvEppParameters(
            world=world_parameters,
            task=task_parameters,
            shared=shared_parameters,
            reference_store=reference_parameters,
            simulator_reset=sim_parameters,
        )

    @field_validator("epp_registry", mode="before")
    def construct_epp_registry_if_necessary_and_validate(cls, epp_registry, values):
        """
        validates the Episode Parameter provider registry
        """
        if epp_registry is None:
            assert (
                "episode_parameter_provider" in values.data
            ), "The validator for `episode_parameter_provider` encountered an exception and did not get passed to this validator"
            assert (
                "episode_parameter_provider_parameters" in values.data
            ), "The validator for `episode_parameter_provider_parameters` encountered an exception and did not get passed to this validator"

            epp_registry = {}
            env_epp_parameters = dict(values.data["episode_parameter_provider_parameters"])
            flat_env_epp_parameters = flatten_dict.flatten(env_epp_parameters)
            env_epp = values.data["episode_parameter_provider"].build(parameters=flat_env_epp_parameters)
            epp_registry[ACT3MultiAgentEnv.episode_parameter_provider_name] = env_epp

            for agent_id, agent_info in values.data["agents"].items():
                agent = agent_info.class_config.agent(
                    agent_name=agent_id, platform_names=agent_info.platform_names, **agent_info.class_config.config
                )
                epp_registry[agent_id] = agent.config.epp

        if ACT3MultiAgentEnv.episode_parameter_provider_name not in epp_registry:
            raise ValueError(f"Missing EPP for '{ACT3MultiAgentEnv.episode_parameter_provider_name}'")

        for agent_id in values.data["agents"]:
            if agent_id not in epp_registry:
                raise ValueError(f"Missing EPP for '{agent_id}'")

        for key, epp in epp_registry.items():
            if not isinstance(epp, EpisodeParameterProvider):
                raise TypeError(
                    f"Invalid type for epp_registry['{key}']: {type(epp)}, only {EpisodeParameterProvider.__qualname__} allowed"
                )

        return epp_registry


class ACT3MultiAgentEnv(BaseCorlMultiAgentEnv):
    """
    ACT3MultiAgentEnv create a RLLIB MultiAgentEnv environment. The following class is intended to wrap
    the interactions with RLLIB and the backend simulator environment. All items here are intended to be
    common parts for running the RLLIB environment with ${simulator} being the unique interaction parts.

    1. Includes wrapping the creation of the simulator specific to run
    2. Includes interactions with the dones, rewards, and glues
    3. etc...
    """

    def __init__(self, config: dict[str, typing.Any]) -> None:  # noqa: PLR0915
        """
        __init__ initializes the rllib multi agent environment

        Parameters
        ----------
        config : ray.rllib.env.env_context.EnvContext
            Passed in configuration for setting items up.
            Must have a 'simulator' key whose value is a BaseIntegrator type
        """
        # this copy protects the callers config from getting corrupted by anything pydantic tries to do
        # really only validators marked with `pre` can do any damage to it
        self.config: ACT3MultiAgentEnvValidator
        super().__init__(config)

        self.seed(1)
        # setup default instance variables
        self._actions: list = []
        self._obs_buffer = ObsBuffer()
        self._reward: dict = {}
        self._done: dict[str, list[DoneFuncBase]] = {}
        self._info: OrderedDict = OrderedDict()
        self._episode_length: int = 0
        self._episode: int = 0
        self._episode_id = None

        # agent glue dict is a mapping from agent id to a dict with keys for the glue names
        # and values of the actual glue object
        self._agent_glue_dict: OrderedDict = OrderedDict()
        self._agent_glue_obs_export_behavior: OrderedDict = OrderedDict()

        # Create the logger
        self._logger = logging.getLogger(ACT3MultiAgentEnv.__name__)

        # Extra simulation init args
        # assign the new output_path with the worker index back to the config for the sim/integration output_path
        extra_sim_init_args: dict[str, typing.Any] = {
            "output_path": str(self.config.output_path),
            "worker_index": self.config.worker_index,
            "vector_index": self.config.vector_index if self.config.vector_index else 0,
        }

        self._agent_dict, extra_sim_init_args["agent_configs"] = env_creation.create_agent_sim_configs(
            self.config.agents,
            self.config.agent_platforms,
            self.config.simulator.type,
            self.config.epp_registry,
            multiple_workers=(self.config.num_workers > 0),
        )
        self._trainable_agent_dict = {k: v for (k, v) in self._agent_dict.items() if v.trainable}
        self._non_trainable_agent_dict = {k: v for (k, v) in self._agent_dict.items() if not (v.trainable)}

        # generate a mapping of agents to platforms and platforms to agents
        # for quick access
        self.agent_to_platforms: dict[str, list[str]] = {}
        self.platform_to_agents: dict[str, list[str]] = {}

        for agent_name, agent_class in self._agent_dict.items():
            self.agent_to_platforms[agent_name] = []
            for platform_name in agent_class.platform_names:
                self.agent_to_platforms[agent_name].append(platform_name)
                if platform_name in self.platform_to_agents:
                    self.platform_to_agents[platform_name].append(agent_name)
                else:
                    self.platform_to_agents[platform_name] = [agent_name]

        # generate a dictionary that keeps track of which platforms that an
        # agent class controls are still operable, when all platforms are inoperable
        # the agent_class will be marked done
        self.agent_platform_inoperable_status = {}
        for agent_name, agent_platforms in self.agent_to_platforms.items():
            self.agent_platform_inoperable_status[agent_name] = {platform_name: False for platform_name in agent_platforms}

        def compute_lcm(values: list[fractions.Fraction]) -> float:
            assert len(values) > 0
            lcm = values[0].denominator
            for v in values:
                lcm = lcm // math.gcd(lcm, v.denominator) * v.denominator
            if lcm == 1:
                lcm = values[0].numerator
                for v in values:
                    lcm = math.gcd(lcm, v.numerator)
                lcm = 1 / lcm  # type: ignore
            return 1.0 / lcm

        max_rate = self.config.max_agent_rate
        self._agent_periods = {
            agent_id: fractions.Fraction(1.0 / agent.frame_rate).limit_denominator(max_rate) for agent_id, agent in self._agent_dict.items()
        }
        self._agent_process_time: dict[str, float] = defaultdict(lambda: sys.float_info.min)
        self.sim_period = compute_lcm(list(self._agent_periods.values()))
        extra_sim_init_args["frame_rate"] = 1.0 / self.sim_period

        # Debug logging
        self._logger.debug(f"output_path : {self.config.output_path}")

        # Sample parameter provider
        default_parameters = self.config.epp.config.parameters

        # Generate topological graph
        # The topological graph only guarantees order necessary to meet dependencies.  Otherwise, the order
        # is unchanged.  If the order of self.config.epp.config.parameters is not constant, the order of
        # random number evaluation can change.  On average, this should still provide good training experiences
        # for the agent; however, it can complicate repeatability studies.  Therefore, set the input of the
        # topological graph to a constant ordering by sorting the default_parameters.
        topological_graph = {}
        for param_path in sorted(default_parameters):
            param = default_parameters[param_path]
            topological_graph[param_path] = set(param.config.dependent_parameters.values())

        self.epp_param_order = tuple(TopologicalSorter(topological_graph).static_order())
        flat_variable_store: dict[tuple[str, ...], typing.Any] = {}

        for k in self.epp_param_order:
            v = default_parameters[k]
            param_value = v.get_value(self.rng, flat_variable_store)
            if isinstance(param_value, dict):
                for param_k, param_v in param_value.items():
                    # for the dict case the param has a path that would be something like
                    # ('simulator_reset', 'platforms', 'paddle0', 'y_and_vel')
                    # we want to collapse down to the nested params so
                    # ('simulator_reset', 'platforms', 'paddle0', 'y')
                    # and ('simulator_reset', 'platforms', 'paddle0', 'vel')
                    # this may be evil and need to be reworked, but not 100% sure
                    flat_variable_store[(*k[:-1], param_k)] = param_v
            else:
                flat_variable_store[k] = param_value

        self._local_variable_store: dict[str, typing.Any] = flatten_dict.unflatten(flat_variable_store)
        for agent in self._agent_dict.values():
            agent.fill_parameters(rng=self.rng, default_parameters=True)

        # !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        # DO NOT CHANGE THIS>>>> STUPID ISSUE CAUSED BY DYNAMIC RESETS OF ENV
        # - RLLIB DOES NOT ALLOW US TO RECREATE THE ENV
        # - We do not want multiple envs
        # ------------>BEGIN<------------
        # !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        if not hasattr(self, "_simulator"):
            self._simulator: BaseSimulator = self.config.simulator.build(**extra_sim_init_args)
        # !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        # ------------>END<------------
        # !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

        self._state, self._sim_reset_args = self._reset_simulator(extra_sim_init_args["agent_configs"])

        # Make the glue objects from the glue mapping now that we have a simulator created
        self._make_glues()

        # create dictionary to hold done history
        self.__setup_state_history()

        # Create the observation and action space now that we have the glue
        # action space operations
        self._action_prop = DictProp(
            spaces=self.__agent_aggregator(agent_list=self._trainable_agent_dict.keys(), agent_function=lambda agent, _: agent.action_prop)
        )
        self._raw_action_space = self._action_prop.create_space()
        gymnasium_space_sort(self._raw_action_space)
        self._action_units = self._action_prop.get_units()
        self.action_space = gymnasium.spaces.Dict(
            self.__agent_aggregator(
                agent_list=self._trainable_agent_dict.keys(), agent_function=lambda agent, _: agent.normalized_action_space
            )
        )
        gymnasium_space_sort(self.action_space)
        self.full_action_space = gymnasium.spaces.Dict(
            self.__agent_aggregator(agent_list=self.agent_dict.keys(), agent_function=lambda agent, _: agent.normalized_action_space)
        )
        gymnasium_space_sort(self.full_action_space)

        # obs space operations
        self._observation_prop = DictProp(
            spaces=self.__agent_aggregator(
                agent_list=self._trainable_agent_dict.keys(), agent_function=lambda agent, _: agent.observation_prop
            )
        )
        self._raw_observation_space = self._observation_prop.create_space()
        self._observation_units = self._observation_prop.get_units()
        self.observation_space = gymnasium.spaces.Dict(
            self.__agent_aggregator(
                agent_list=self._trainable_agent_dict.keys(), agent_function=lambda agent, _: agent.normalized_observation_space
            )
        )
        self.full_observation_space = gymnasium.spaces.Dict(
            self.__agent_aggregator(agent_list=self.agent_dict.keys(), agent_function=lambda agent, _: agent.normalized_observation_space)
        )

        self._set_space_seeds()

        # full obs
        self._full_observation_prop = DictProp(
            spaces=self.__agent_aggregator(agent_list=self.agent_dict.keys(), agent_function=lambda agent, _: agent.observation_prop)
        )
        self._full_observation_units = self._full_observation_prop.get_units()

        self._shared_done: DoneDict = DoneDict()
        self._done_info: OrderedDict = OrderedDict()
        self._reward_info: OrderedDict = OrderedDict()

        self._episode_init_params: dict

        self.done_string = ""
        self._agent_ids = set(self._raw_action_space.spaces.keys())

        self._skip_action = False
        self._agent_ids = list(self._agent_dict.keys())
        self._agent_ids_trainable = list(self._trainable_agent_dict.keys())
        self._agent_ids_non_trainable = list(self._non_trainable_agent_dict.keys())
        self.action_space_in_preferred_format = True
        self._obs_space_in_preferred_format = True

    @staticmethod
    def get_validator() -> type[ACT3MultiAgentEnvValidator]:
        """Get the validator for this class."""
        return ACT3MultiAgentEnvValidator

    def reset(self, *, seed=None, options=None) -> tuple[OrderedDict, OrderedDict]:
        if seed is not None:
            self.seed(seed)
        # Sample parameter provider
        current_parameters, self._episode_id, env_epp_ctx = self.config.epp.get_params(self.rng)

        flat_variable_store: dict[tuple[str, ...], typing.Any] = {}
        for k in self.epp_param_order:
            v = current_parameters[k]
            param_value = v.get_value(self.rng, flat_variable_store)
            if isinstance(param_value, dict):
                for param_k, param_v in param_value.items():
                    # for the dict case the param has a path that would be something like
                    # ('simulator_reset', 'platforms', 'paddle0', 'y_and_vel')
                    # we want to collapse down to the nested params so
                    # ('simulator_reset', 'platforms', 'paddle0', 'y')
                    # and ('simulator_reset', 'platforms', 'paddle0', 'vel')
                    # this may be evil and need to be reworked, but not 100% sure
                    flat_variable_store[(*k[:-1], param_k)] = param_v
            else:
                flat_variable_store[k] = param_value
        self._local_variable_store = flatten_dict.unflatten(flat_variable_store)

        # Reset all agents. This should be done prior to calling other
        # functionality on the agents in case the agents' reset()
        # changes agent attributes used to generate items like variable
        # store parameters, rewards, etc.
        for agent in self._agent_dict.values():
            agent.reset()

        for agent in self._agent_dict.values():
            agent.fill_parameters(self.rng, env_epp_ctx)

        # 3. Reset the Done and Reward dictionaries for the next iteration
        self._make_rewards()
        self._done = {}
        self._done_info.clear()
        self._make_dones()
        self._shared_done = self._make_shared_dones()

        self._reward = {}

        # generate a dictionary that keeps track of which platforms that an
        # agent class controls are still operable, when all platforms are inoperable
        # the agent_class will be marked done
        self.agent_platform_inoperable_status = {}
        for agent_name, agent_platforms in self.agent_to_platforms.items():
            self.agent_platform_inoperable_status[agent_name] = {platform_name: False for platform_name in agent_platforms}

        self._set_default_done_reward()
        # 4. Reset the simulation/integration
        self._state, self._sim_reset_args = self._reset_simulator()
        self._episode_length = 0
        self._actions.clear()
        self._episode += 1

        self._agent_process_time.clear()

        #####################################################################
        # Make glue sections - Given the state of the simulation we need to
        # update the platform interfaces.
        #####################################################################
        self._make_glues()

        #####################################################################
        # get observations
        # For each configured agent read the observations/measurements
        #####################################################################
        agent_list = list(self._agent_dict.keys())
        # The following loop guarantees that during training that the glue
        # states start with valid values for rates. The number of recommended
        # steps for sim is at least equal to the depth of the rate observation
        # tree (ex: speed - 2, acceleration - 3, jerk - 4) - recommend defaulting
        # to 4 as we do not go higher thank jerk
        # 1 step is always added for the initial obs in reset
        # 1Hz 10 Seconds = 10 --- 2Hz 10 seconds = 20 --- .5 Hz 10 Seconds = 5
        warmup_steps = math.ceil(self.config.sim_warmup_seconds * self._state.sim_update_rate_hz)
        platforms: set[str] = set()
        for _ in range(warmup_steps):
            self._state = self._simulator.step(platforms)
        self._obs_buffer.next_observation = self.__get_observations_from_glues(agent_list)
        self._obs_buffer.update_obs_pointer()
        self.__setup_state_history()

        for platform_name in self._state.sim_platforms:
            # self._state.episode_history[platform_name].clear()
            self._state.episode_state[platform_name] = OrderedDict()
        for agent_name in self._agent_dict:
            self._state.agent_episode_state[agent_name] = OrderedDict()

        self._create_actions(self._agent_dict, self._obs_buffer.observation)

        self._info.clear()
        self.__get_info_from_glue(agent_list)
        trainable_info = self._get_trainable_info(agent_list)

        training_obs = self._create_training_observations(agent_list, self._obs_buffer)

        return training_obs, trainable_info

    def _reset_simulator(self, agent_configs=None) -> tuple[BaseSimulatorState, dict[str, typing.Any]]:
        sim_reset_args = copy.deepcopy(self.config.simulator_reset_parameters)
        v_store = self._local_variable_store
        deepmerge.always_merger.merge(sim_reset_args, v_store.get("simulator_reset", {}))

        self._process_references(sim_reset_args, v_store)

        for agent_data in self._agent_dict.values():
            deepmerge.always_merger.merge(
                sim_reset_args,
                {"platforms": {platform_name: agent_data.get_simulator_reset_parameters() for platform_name in agent_data.platform_names}},
            )

        if agent_configs is not None:
            sim_reset_args["agent_configs_reset"] = agent_configs

        return self._simulator.reset(sim_reset_args), sim_reset_args

    def _process_references(self, sim_reset_args: dict, v_store: dict) -> None:
        """Process the reference store look ups for the position data

        Parameters
        ----------
        sim_reset_args : dict
            The simulator reset parameters
        v_store : dict
            The variable store
        """
        plat_str = "platforms"
        if plat_str in sim_reset_args:
            pos_str = "position"
            ref_str = "reference"
            for plat_k, plat_v in sim_reset_args[plat_str].items():
                if pos_str in plat_v:
                    for position_k, position_v in plat_v[pos_str].items():
                        if isinstance(position_v, dict) and ref_str in position_v:
                            sim_reset_args[plat_str][plat_k][pos_str][position_k] = v_store["reference_store"].get(
                                position_v[ref_str], self.config.reference_store[position_v[ref_str]]
                            )

    def _remove_non_trainable_agent_ids(self, agent_list: typing.Iterable[str]) -> list[str]:
        """
        Remove any agent IDs corresponding to non-trainable agents.
        """

        return [a for a in agent_list if a not in self._agent_ids_non_trainable]

    def _get_operable_agent_platforms(self):
        return {
            name
            for name, item in self.state.sim_platforms.items()
            if item.operable and not self.state.episode_state.get(name, {}) and self.platform_to_agents.get(name, None)
        }

    def _get_operable_agents(self):
        """Determines which agents are operable in the sim, this becomes stale after the simulation is stepped"""

        operable_agents = []
        for agent_name, platform_inoperability_status in self.agent_platform_inoperable_status.items():
            inoperability_values = platform_inoperability_status.values()

            if any(inoperability_values):
                self._agent_dict[agent_name].set_removed(platform_inoperability_status)
            if not all(inoperability_values):
                operable_agents.append(agent_name)

        return operable_agents

    def _get_trainable_info(self, agent_list: typing.Collection[str]) -> typing.OrderedDict:
        """Returns the training info"""

        agent_list = self._remove_non_trainable_agent_ids(agent_list)

        trainable_info = get_dictionary_subset(self._info, agent_list)

        # add platform obs and env data to trainable_info (for use by custom policies)
        for agent_id in agent_list:
            if agent_id not in trainable_info:
                trainable_info[agent_id] = {}

            trainable_info[agent_id]["env"] = {"sim_period": self.sim_period, "sim_time": self._state.sim_time}

            platform_indices = {platform_id: idx for idx, platform_id in enumerate(self.platform_to_agents)}
            agents_to_platform_indices = {
                agent: [platform_indices[platform] for platform in platforms] for agent, platforms in self.agent_to_platforms.items()
            }

            trainable_info[agent_id]["platform_to_index"] = platform_indices
            trainable_info[agent_id]["agents_to_platform_indices"] = agents_to_platform_indices

        return trainable_info

    def step(self, action_dict: dict):  # noqa: PLR0915,PLR0912
        """Returns observations from ready agents.

        The returns are dicts mapping from agent_id strings to values. The
        number of agents in the env can vary over time.

            obs (StateDict): New observations for each ready agent.
                episode is just started, the value will be None.
            dones (StateDict): Done values for each ready agent. The
                special key "__all__" (required) is used to indicate env
                termination.
            infos (StateDict): Optional info values for each agent id.
        """
        self._episode_length += 1

        operable_platforms = self._get_operable_agent_platforms()
        operable_agents = self._get_operable_agents()

        for agent_id in self._non_trainable_agent_dict:
            _observation = self._obs_buffer.observation.get(agent_id)
            if _observation is None:
                continue
            _action = self._non_trainable_agent_dict[agent_id].create_action(observation=_observation)  # type: ignore
            action_dict[agent_id] = _action

        # look to add this bugsplat, but this check won't work for multi fps
        # if set(operable_agents.keys()) != set(action_dict.keys()):
        #     raise RuntimeError("Operable_agents and action_dict keys differ!"
        #                        f"operable={set(operable_agents.keys())} != act={set(action_dict.keys())} "
        #                        "If this happens that means either your platform is not setting non operable correctly"
        #                        " (if extra keys are in operable set) or you do not have a done condition covering "
        #                        "a condition where your platform is going non operable. (if extra keys in act)")
        raw_action_dict = {} if self._skip_action else self._apply_action(operable_agents, action_dict)

        # Save current action for future debugging
        if not self.config.disable_action_history:
            self._actions.append(action_dict)

        platforms_to_process = set()
        for agent_id in action_dict:
            for platform_name in self.agent_to_platforms[agent_id]:
                if platform_name in operable_platforms:
                    platforms_to_process.add(platform_name)

        try:
            self._state = self._simulator.step(platforms_to_process)
        except ValueError as err:
            if self.config.save_state_pickle:
                self._save_state_pickle(err)

        # MTB - Changing to not replace operable_agents variable
        #       We calculate observations on agents operable after sim step
        #       - This is done because otherwise observations would be invalid
        #       Calculate Dones/Rewards on agents operable before sim step
        #       - This is done because if an agent "dies" it needs to have a final done calculated
        operable_platforms_after_step = self._get_operable_agent_platforms()
        # get the difference in operable platforms to find those that are inoperable
        # then we mark it inoperable so that _get_operable_agents can correctly do it's job
        new_inoperable_platforms = operable_platforms - operable_platforms_after_step
        for platform_name in new_inoperable_platforms:
            for agent_name in self.platform_to_agents[platform_name]:
                self.agent_platform_inoperable_status[agent_name][platform_name] = True

        operable_agents_after_step = self._get_operable_agents()

        #####################################################################
        # get next observations - For each configured platform read the
        # observations/measurements
        #####################################################################
        self._obs_buffer.next_observation = self.__get_observations_from_glues(operable_agents_after_step)

        self._info.clear()
        self.__get_info_from_glue(operable_agents_after_step)

        #####################################################################
        # Process the done conditions
        # 1. Reset the rewards from the last step
        # 2. loops over all agents and processes the reward conditions per
        #    agent
        #####################################################################

        platforms_done = self.__get_done_from_platforms(operable_platforms, raw_action_dict=raw_action_dict)

        expected_done_keys = operable_platforms
        if set(platforms_done.keys()) != expected_done_keys:
            raise RuntimeError(
                f'Local dones do not match expected keys.  Received "{platforms_done.keys()}".  Expected "{expected_done_keys}".'
            )

        shared_dones, shared_done_info = self._shared_done(
            observation=self._obs_buffer.observation,
            action=raw_action_dict,
            next_observation=self._obs_buffer.next_observation,
            next_state=self._state,
            observation_space=self._raw_observation_space,
            observation_units=self._observation_units,
            local_dones=copy.deepcopy(platforms_done),
            local_done_info=copy.deepcopy(self._done_info),
        )

        if shared_dones.keys():
            if set(shared_dones.keys()) != expected_done_keys:
                raise RuntimeError(
                    f'Shared dones do not match expected keys.  Received "{shared_dones.keys()}".  Expected "{expected_done_keys}".'
                )
            for key in expected_done_keys:
                platforms_done[key] |= shared_dones[key]

            assert shared_done_info is not None

            local_done_info_keys = set(self._done_info.keys())
            shared_done_info_keys = set(shared_done_info)
            common_keys = local_done_info_keys & shared_done_info_keys
            if not common_keys:
                raise RuntimeError(f'Dones do not have common keys: "{common_keys}"')

            for platform_name in filter(lambda x: x in shared_done_info, self._done_info):
                self._done_info[platform_name].update(shared_done_info[platform_name])

        # we now have all platforms that are done
        # convert these platform done's to agent dones
        # update state to map platform episode state to agent episode state
        for platform_name, _ in filter(lambda x: x[1], platforms_done.items()):
            for agent_name in self.platform_to_agents[platform_name]:
                self.agent_platform_inoperable_status[agent_name][platform_name] = True
                for done_name, done_status in self._state.episode_state[platform_name].items():
                    self._state.agent_episode_state[agent_name][done_name] = done_status

        # compute agent dones from the platform dones
        # update episode state to hold information about agent done status
        agents_done = {}
        for agent_name, agent_platform_dones in self.agent_platform_inoperable_status.items():
            agents_done[agent_name] = all(agent_platform_dones.values())

        # calculate if all agents are done
        if self.config.end_episode_on_first_agent_done:
            agents_done["__all__"] = any(agents_done.values())
        else:
            agents_done["__all__"] = all(agents_done.values())

        # Update agent_episode_state based on if shared dones modified episode_state after an agent was done
        if agents_done["__all__"]:
            for platform_name in self.state.episode_state:
                for agent_name in self.platform_to_agents.get(platform_name, []):
                    for done_name, done_status in self._state.episode_state[platform_name].items():
                        self._state.agent_episode_state[agent_name][done_name] = done_status

        # Tell the simulator to mark the episode complete
        if agents_done["__all__"]:
            self._simulator.mark_episode_done(self._done_info, self._state.episode_state, self.config.metadata)

        self._reward.clear()

        if agents_done["__all__"]:
            agents_to_process_this_timestep = operable_agents
        else:

            def do_process_agent(self, agent_id) -> bool:
                frame_rate = self._agent_periods[agent_id].numerator / self._agent_periods[agent_id].denominator
                return self._state.sim_time >= self._agent_process_time[agent_id] + frame_rate - self.config.timestep_epsilon

            agents_to_process_this_timestep = list(filter(partial(do_process_agent, self), operable_agents))

        for agent_id in agents_to_process_this_timestep:
            self._agent_process_time[agent_id] = self._state.sim_time

        reward = self.__get_reward_from_agents(agents_to_process_this_timestep, raw_action_dict=raw_action_dict)

        self._simulator.save_episode_information(
            self.done_info, self.reward_info, self._obs_buffer.next_observation, self._full_observation_units
        )
        # copy over observation from next to previous - There is no real reason to deep
        # copy here. The process of getting a new observation from the glue copies. All
        # we need to do is maintain the order of two buffers!!!.
        # Tested with: They are different and decreasing as expected
        #   print(f"C: {self._obs_buffer.observation['blue0']['ObserveSensor_Sensor_Fuel']}")
        #   print(f"N: {self._obs_buffer.next_observation['blue0']['ObserveSensor_Sensor_Fuel']}")
        self._obs_buffer.update_obs_pointer()

        # Note that `_create_training_observations` and `_get_trainable_info`
        # contain a filter for trainable agents.
        trainable_observations = self._create_training_observations(agents_to_process_this_timestep, self._obs_buffer)
        trainable_info = self._get_trainable_info(agents_to_process_this_timestep)

        # Filters 'agent IDs to process this timestep' so that it only
        # includes IDs of trainable agents.
        trainable_agents_to_process_this_timestep = self._remove_non_trainable_agent_ids(agents_to_process_this_timestep)
        trainable_rewards = get_dictionary_subset(reward, trainable_agents_to_process_this_timestep)
        trainable_dones = get_dictionary_subset(agents_done, [*list(trainable_agents_to_process_this_timestep), "__all__"])

        # if not done all, delete any platforms from simulation that are done, so they don't interfere
        # go through done's and update inoperable platform list so that operable platforms can be updated
        if not agents_done["__all__"]:
            platforms_deleted = set()
            for platform_name, platform_done in platforms_done.items():
                if platform_done:
                    self.simulator.delete_platform(platform_name)
                    platforms_deleted.add(platform_name)

        #####################################################################
        # return results to RLLIB - Note that RLLIB does not do a recursive
        # isinstance call and as such need to make sure items are
        # OrderedDicts
        #####################################################################
        return trainable_observations, trainable_rewards, trainable_dones, {"__all__": trainable_dones["__all__"]}, trainable_info

    def __get_done_from_platforms(self, alive_platforms: typing.Iterable[str], raw_action_dict):
        def or_merge(config, path, base, nxt):
            return base or nxt

        merge_strategies = copy.deepcopy(deepmerge.DEFAULT_TYPE_SPECIFIC_MERGE_STRATEGIES)
        merge_strategies.append((bool, or_merge))
        or_merger = deepmerge.Merger(merge_strategies, [], [])

        done = OrderedDict()

        for platform_name in filter(lambda x: x in self._done, alive_platforms):
            platform_is_done = False
            done_info = {}
            for platform_done in self._done[platform_name]:
                ret = platform_done(
                    observation=self._obs_buffer.observation,
                    action=raw_action_dict,
                    next_observation=self._obs_buffer.next_observation,
                    next_state=self._state,
                    observation_space=self._raw_observation_space,
                    observation_units=self._observation_units,
                )
                platform_is_done = platform_is_done or ret
                done_info[platform_done.name] = ret

            done[platform_name] = platform_is_done
            # get around reduction
            or_merger.merge(self._done_info.setdefault(platform_name, {}), done_info)

            self._done_info[platform_name] = done_info

        return done

    def __get_reward_from_agents(self, alive_agents: typing.Iterable[str], raw_action_dict):
        reward = OrderedDict()
        for agent_id in alive_agents:
            agent_class = self._agent_dict[agent_id]
            agent_reward, reward_info = agent_class.get_rewards(
                observation=self._obs_buffer.observation,
                action=raw_action_dict,
                next_observation=self._obs_buffer.next_observation,
                state=self._state,
                next_state=self._state,
                observation_space=self._raw_observation_space,
                observation_units=self._observation_units,
            )
            # it is possible to have a HL policy that does not compute an reward
            # in this case just return a zero for reward value
            reward[agent_id] = agent_reward
            self._reward_info[agent_id] = reward_info
        return reward

    def _set_default_done_reward(self):
        """
        Populate the done/rewards with default values
        """
        for key in self._agent_dict:
            self._reward[key] = 0
            self._shared_done[key] = False
        self._shared_done[SharedDoneFuncBase.ALL] = False

    def _create_training_observations(self, alive_agents: typing.Iterable[str], observations: ObsBuffer) -> OrderedDict:
        """
        Filters and normalizes observations (the sample of the space) using the glue normalize functions.

        Parameters
        ----------
        alive_agents:
            The agents that are still alive
        observations:
            The observations

        Returns
        -------
        OrderedDict:
            the filtered/normalized observation samples
        """

        alive_agents = self._remove_non_trainable_agent_ids(alive_agents)

        this_steps_obs = OrderedDict()
        for agent_id in alive_agents:
            if agent_id in observations.observation:
                this_steps_obs[agent_id] = observations.observation[agent_id]
            elif agent_id in observations.next_observation:
                this_steps_obs[agent_id] = observations.next_observation[agent_id]
            else:
                raise RuntimeError(
                    "ERROR: _create_training_observations tried to retrieve obs for this training step"
                    f" but {agent_id=} was not able to be found in either the current obs data or the "
                    " obs from the previous timestep as a fallback"
                )

        # sanity check will fail on quantities, and we definitely don't want to send quantities into an RL
        # framework
        unit_extracted_obs = tree.map_structure(lambda v: v.m if isinstance(v, Quantity) else v, this_steps_obs)
        # Sanity checks and Scale - ensure run first time and run only every N times...
        # Same as RLLIB - This can add a bit of time as we are exploring complex dictionaries
        # default to every time if not specified... Once the limits are good we it is
        # recommended to increase this for training
        if self.config.deep_sanity_check and self._episode_length % self.config.sanity_check_obs == 0:
            try:
                self._sanity_check(self._raw_observation_space, unit_extracted_obs)
            except ValueError as err:
                if self.config.save_state_pickle:
                    self._save_state_pickle(err)

        return OrderedDict(
            self.__agent_aggregator(
                agent_list=alive_agents,
                agent_function=lambda agent, agent_id, all_agent_obs: agent.create_training_observations(all_agent_obs[agent_id]),
                all_agent_obs=unit_extracted_obs,
            )
        )

    def __get_observations_from_glues(self, alive_agents: typing.Iterable[str]) -> OrderedDict:
        """
        Gets the observation dict from all the glue objects for each agent

        Returns
        -------
        OrderedDict:
            The observation dict from all the glues
        """
        return_observation: OrderedDict = OrderedDict()
        for agent_id in alive_agents:
            agent_class = self._agent_dict[agent_id]

            glue_obj_obs = agent_class.get_observations()
            if len(glue_obj_obs) > 0:
                return_observation[agent_id] = glue_obj_obs
        return return_observation

    def _apply_action(self, operable_agents, action_dict):
        raw_action_dict = OrderedDict()
        for agent_id in operable_agents:
            agent_class = self._agent_dict[agent_id]
            if agent_id in action_dict:
                raw_action_dict[agent_id] = agent_class.apply_action(action_dict[agent_id], self._obs_buffer.observation[agent_id])
        return raw_action_dict

    def __get_info_from_glue(self, alive_agents: typing.Iterable[str]):
        for agent_id in alive_agents:
            agent_class = self._agent_dict[agent_id]
            glue_obj_info = agent_class.get_info_dict()
            if len(glue_obj_info) > 0:
                self._info[agent_id] = glue_obj_info

    def _make_glues(self) -> None:
        """ """
        env_ref_stores = [self._local_variable_store.get("reference_store", {}), self.config.reference_store]
        for agent, agent_class in self._agent_dict.items():
            agent_plats = {platform_name: self._get_platform_by_name(platform_name) for platform_name in agent_class.platform_names}
            agent_class.make_glues(agent_plats, agent, env_ref_stores=env_ref_stores)

    def _make_rewards(self) -> None:
        """ """
        env_ref_stores = [self._local_variable_store.get("reference_store", {}), self.config.reference_store]

        for agent, agent_class in self._agent_dict.items():
            agent_class.make_rewards(agent, env_ref_stores=env_ref_stores)

    def _make_dones(self) -> None:
        """
        register done functors to platforms, but done conditions are assigned to a specific platform
        """
        env_ref_stores = [self._local_variable_store.get("reference_store", {}), self.config.reference_store]
        # 1Hz 10 Seconds = 10 --- 2Hz 10 seconds = 20 --- .5 Hz 10 Seconds = 5
        warmup_steps = math.ceil(self.config.sim_warmup_seconds * self._state.sim_update_rate_hz)
        episode_length_done = Functor(
            functor=EpisodeLengthDone,
            config={"horizon": corl_quantity()(value=(self.config.horizon + warmup_steps) * self.sim_period, units="sec")},
        )

        platform_dones_collection: dict[str, list[DoneFuncBase]] = {}
        platform_done_dicts = {}
        for agent_class in self._agent_dict.values():
            env_params = {
                platform_name: [
                    self._local_variable_store.get("world", {}),
                    self._local_variable_store.get("task", {}).get(platform_name, {}),
                ]
                for platform_name in agent_class.platform_names
            }
            tmp = agent_class.make_platform_dones(env_params=env_params, env_ref_stores=env_ref_stores)
            for platform_name, platform_dones in tmp.items():
                if platform_name in platform_dones_collection:
                    platform_dones_collection[platform_name].extend(platform_dones)
                else:
                    platform_dones_collection[platform_name] = platform_dones

        for platform_name in self.platform_to_agents:
            platform_dones = platform_dones_collection.get(platform_name, [])
            for done_dict in chain(self.config.dones.world, self.config.dones.task[platform_name], [episode_length_done]):
                platform_dones.append(
                    done_dict.create_functor_object(
                        param_sources=[
                            self._local_variable_store.get("world", {}),
                            self._local_variable_store.get("task", {}).get(platform_name, {}),
                        ],
                        ref_sources=env_ref_stores,
                        platform_name=platform_name,
                    )
                )
            # platform_done_dict = DoneDict(processing_funcs=platform_dones)  # type: ignore
            platform_done_dicts[platform_name] = platform_dones

        self._done = platform_done_dicts

    def _make_shared_dones(self) -> DoneDict:
        """
        _get_shared_done_functors gets and initializes the
        shared done dict used for this iteration

        The shared done dictionary does not correspond to individual
        agents but looks sharedly at all agents.

        this will be called after any updates to the simulator
        configuration during reset

        Returns
        -------
        DoneDict
            The DoneDict with functors used for this iteration
        """
        done_conditions = []
        ref_sources = [self._local_variable_store.get("reference_store", {}), self.config.reference_store]
        param_sources = [self._local_variable_store.get("shared", {})]
        for done_functor in self.config.dones.shared:
            tmp = done_functor.create_functor_object(param_sources=param_sources, ref_sources=ref_sources)
            done_conditions.append(tmp)
        return DoneDict(processing_funcs=done_conditions)

    @staticmethod
    def _create_actions(
        agents,
        observations: dict | None = None,
        rewards: dict | None = None,
        dones: dict | None = None,
        info: dict | None = None,
    ):
        if info is None:
            info = dict()
        if dones is None:
            dones = dict()
        if rewards is None:
            rewards = dict()
        if observations is None:
            observations = dict()
        for agent_name, agent_class in agents.items():
            agent_class.create_next_action(
                observations.get(agent_name), rewards.get(agent_name), dones.get(agent_name), info.get(agent_name)
            )

    def __agent_aggregator(self, agent_list, agent_function: typing.Callable, **kwargs) -> dict:
        """
        __agent_aggregator calls function on all agents and places the output in the in a dictionary.
        Dictionary key set to the agent name and values are set to the return value of the method

        Parameters
        ----------
        agent_function
            A function that takes an agent: BaseAgent as an argument
        Returns
        -------
            A dictionary of agent names to method output
        """
        # init our return Dict
        agg_dict = {}
        # loop over all the agents and their glue_name_obj_pairs list
        for agent_id in agent_list:
            agent_class = self._agent_dict[agent_id]
            agg_dict[agent_id] = agent_function(agent_class, agent_id, **kwargs)
            # if this agent provided anything add it to the return space
        return agg_dict

    @property
    def state(self) -> BaseSimulatorState:
        """
        state of platform object.  Current state.

        Returns
        -------
        BaseSimulatorState
            the dict storing the current state of environment.
        """
        return self._state

    @property
    def simulator(self) -> BaseSimulator:
        """
        simulator simulator instance

        Returns
        -------
        BaseSimulator
            The simulator instance in the base
        """
        return self._simulator

    @property
    def actions(self) -> list[typing.Any]:
        """
        get the current episode parameter provider episode id

        Returns
        -------
        list
            actions
        """
        return self._actions

    @property
    def observation(self) -> OrderedDict:
        """
        observation get the observation for the agents in this environment

        Returns
        -------
        OrderedDict
            the dict holding the observations for the agents
        """
        return self._obs_buffer.observation

    @property
    def episode_id(self) -> int | None:
        """
        get the current episode parameter provider episode id

        Returns
        -------
        int or None
            the episode id
        """
        return self._episode_id

    @property
    def local_variable_store(self) -> dict[str, typing.Any]:
        """
        get the local variable store

        Returns
        -------
        Dict
            the local variable store episode id
        """
        return self._local_variable_store

    @property
    def epp(self) -> EpisodeParameterProvider:
        """
        get the Episode Parameter Provider

        Returns
        -------
        EpisodeParameterProvider
        """
        return self.config.epp

    @property
    def agent_dict(self) -> dict[str, typing.Any]:
        """
        get the full agent dict

        Returns
        -------
        dict[str, typing.Any]
            A dictionary of {agent_id, agent} for agents in env.
        """
        return self._agent_dict

    def _get_platform_by_name(self, platform_id: str) -> BasePlatform:
        platform: BasePlatform = self._state.sim_platforms.get(platform_id, None)

        if platform is None or not issubclass(platform.__class__, BasePlatform):
            self._logger.error("-" * 100)
            self._logger.error(f"{self._state}")
            for i in self._state.sim_platforms:
                self._logger.error(f"{i}")
            raise ValueError(f"{self.__class__.__name__} glue could not find a platform named {platform_id} of class BasePlatform")

        return platform

    def __setup_state_history(self):
        # self._state.episode_history = defaultdict(partial(deque, maxlen=self.config.horizon))
        pass

    def post_process_trajectory(self, agent_id, batch, episode, policy):
        """easy accessor for calling post process trajectory
            correctly

        Arguments:
            agent_id: agent id
            batch: post processed Batch - be careful modifying
        """
        self._agent_dict[agent_id].post_process_trajectory(
            agent_id, episode.worker.env._state, batch, episode, policy, episode.worker.env._reward_info  # noqa: SLF001  # noqa: SLF001
        )

    @staticmethod
    def _sanity_check(space: gymnasium.spaces.Space, space_sample: EnvSpaceUtil.sample_type) -> None:
        """
        Sanity checks a space_sample against a space
        1. Check to ensure that the sample from the integration base
           Fall within the expected range of values.

        Note: space_sample and space expected to match up on
        Key level entries

        Parameters
        ----------
        space: gymnasium.spaces.Space
            the space to check the sample against
        space_sample: EnvSpaceUtil.sample_type
            the sample to check if it is actually in the bounds of the space

        Returns
        -------
        OrderedDict:
            the scaled_observations
        """
        if not space.contains(space_sample):
            EnvSpaceUtil.deep_sanity_check_space_sample(space, space_sample)

    def _save_state_pickle(self, err: ValueError):
        """saves state for later debug

        Arguments:
            err {ValueError} -- Traceback for the error

        Raises:
            err: Customized error message to raise for the exception

        """
        out_pickle = str(Path(self.config.output_path) / f"sanity_check_failure_{self._episode}.pkl")
        p_dict: dict[str, typing.Any] = {}
        p_dict["err"] = str(err)

        class NumpyArrayEncoder(JSONEncoder):
            """Encode the numpy types for json"""

            def default(self, o):
                val = None
                if isinstance(
                    o,
                    np.int_ | np.intc | np.intp | np.int8 | np.int16 | np.int32 | np.int64 | np.uint8 | np.uint16 | np.uint32 | np.uint64,
                ):
                    val = int(o)
                elif isinstance(o, np.float_ | np.float16 | np.float32 | np.float64):
                    val = float(o)
                elif isinstance(o, np.ndarray):  # This is the fix
                    val = o.tolist()
                elif isinstance(o, np.bool_):
                    val = "True" if o is True else "False"
                else:
                    val = json.JSONEncoder.default(self, o)

                return val

        def to_dict(input_ordered_dict):
            return loads(dumps(input_ordered_dict, cls=NumpyArrayEncoder))

        p_dict["action"] = self._actions
        p_dict["observation"] = self._obs_buffer.observation
        p_dict["dones"] = to_dict(self._done_info)
        # p_dict["env_config"] = copy.deepcopy(self.env_config)  # type: ignore
        p_dict["step"] = str(self._episode_length)

        with open(out_pickle, "wb") as f:
            pickle.dump(p_dict, f)

        raise ValueError(f"Error occurred: {err} \n Saving sanity check failure output pickle to file: {out_pickle}")

    def seed(self, seed=None):
        """generates environment seed through rllib

        Keyword Arguments:
            seed {[int]} -- seed to set environment with (default: {None})

        Returns:
            [int] -- [seed value]
        """
        self.rng, self.seed_value = gymnasium.utils.seeding.np_random(seed)

        # Call to self._set_space_seeds() will fail during __init__ because these attributes do not exist yet.
        with contextlib.suppress(AttributeError):
            self._set_space_seeds()

        return [self.seed_value]

    def _set_space_seeds(self) -> None:
        self.observation_space.seed(self.seed_value)
        self._raw_observation_space.seed(self.seed_value)
        self.action_space.seed(self.seed_value)
        self._raw_action_space.seed(self.seed_value)



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
"""
from ray.tune.registry import ENV_CREATOR, _global_registry


def get_rllib_environment_creator(env_specifier):
    """
    Returns CoRL environment
    """
    if _global_registry.contains(ENV_CREATOR, env_specifier):
        return _global_registry.get(ENV_CREATOR, env_specifier)

    raise RuntimeError(f"No registered env creator named {env_specifier}")



"""
---------------------------------------------------------------------------

This is a US Government Work not subject to copyright protection in the US.
The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
"""

import os
import pickle
from abc import abstractmethod
from datetime import datetime
from pathlib import Path

import flatten_dict
import gymnasium
import numpy as np
from ray.rllib import BaseEnv, RolloutWorker
from ray.rllib.evaluation.episode_v2 import EpisodeV2
from ray.rllib.models.preprocessors import get_preprocessor
from ray.rllib.policy import Policy
from ray.rllib.policy.sample_batch import SampleBatch
from ray.rllib.utils.typing import AgentID, PolicyID

from corl.environment.multi_agent_env import ACT3MultiAgentEnv
from corl.evaluation.episode_artifact import EpisodeArtifact
from corl.evaluation.runners.section_factories.engine.rllib.default_evaluation_callbacks import DefaultEvaluationCallbacks

EPOCH_ZFILL = 6
EVAL_CHECKPOINTS_SUBDIR = "eval_checkpoints"


def set_shareable_info(worker: RolloutWorker, training_iteration: int):
    """Sets worker attributes

    Parameters
    ----------
    worker : RolloutWorker
        The worker, either evaluation worker or (training) worker
    training_iteration : int
        The training iteration
    """
    if worker.config.in_evaluation:
        worker.config["training_iteration"] = training_iteration


class EpisodeArtifactLoggingCallback(DefaultEvaluationCallbacks):
    """
    This is basically the same callback used in the evaluation to same trajectory rollouts.
    The difference is that writing to processed batch seems to mess up training,
    and we changed the naming of the files and where they are written.
    """

    def __init__(self):
        super().__init__()
        self._epoch = 0
        self._episode_counter = 0

    @abstractmethod
    def platform_serializer(self):
        """
        Method specifying the serialize function
        """
        raise NotImplementedError

    def extract_environment_state(self, env_state):  # noqa: PLR6301
        return {}

    def on_evaluate_start(
        self,  # noqa: PLR6301
        *,
        algorithm,
        **kwargs,
    ) -> None:
        """Callback before evaluation starts.

        This method gets called at the beginning of Algorithm.evaluate().

        Args:
            algorithm: Reference to the algorithm instance.
            kwargs: Forward compatibility placeholder.
        """
        training_iteration = algorithm.training_iteration
        algorithm.evaluation_workers.foreach_worker(lambda worker: set_shareable_info(worker, training_iteration), healthy_only=True)

        # Saves the checkpoint into `eval_checkpoints`
        epoch_str = str(training_iteration).zfill(EPOCH_ZFILL)

        output_path = Path(algorithm._logdir).joinpath(EVAL_CHECKPOINTS_SUBDIR, f"epoch_{epoch_str}")  # noqa: SLF001
        os.makedirs(output_path, exist_ok=True)
        algorithm.save_checkpoint(output_path)

        super().on_evaluate_start(algorithm=algorithm, **kwargs)

    def on_episode_start(
        self,  # noqa: PLR6301
        *,
        worker: RolloutWorker,
        base_env: BaseEnv,
        policies: dict[PolicyID, Policy],
        episode: EpisodeV2,
        **kwargs,
    ) -> None:
        # At this point, episode.user_data has been initialized. Here we add the initial state.
        env = base_env.get_sub_environments()[episode.env_id]
        episode.user_data["initial_state"] = env.observation
        super().on_episode_start(worker=worker, base_env=base_env, policies=policies, episode=episode, **kwargs)

    def on_episode_step(
        self,  # noqa: PLR6301
        *,
        worker: RolloutWorker,
        base_env: BaseEnv,
        policies: dict[PolicyID, Policy] | None = None,
        episode: EpisodeV2,
        **kwargs,
    ) -> None:
        # Logs the magnitude of each action
        env = base_env.get_sub_environments()[episode.env_id]
        if env.actions:
            # Cycles through each agent
            for agent_id, action_dict in env.actions[-1].items():
                # Flattens the dictionary
                flattened_action_dict = flatten_dict.flatten(action_dict, reducer="dot")
                # Cycles each each action
                for flattened_name, value in flattened_action_dict.items():
                    metric_name = f"{agent_id}/{flattened_name}"
                    episode.custom_metrics[metric_name] = value
                    episode.custom_metrics[f"{metric_name}_mag"] = np.abs(value)

        super().on_episode_step(worker=worker, policies=policies, base_env=base_env, episode=episode, **kwargs)

    def on_episode_end(
        self, *, worker: RolloutWorker, base_env: BaseEnv, policies: dict[PolicyID, Policy], episode: EpisodeV2, **kwargs
    ) -> None:
        """
        on_episode_end stores the custom metrics in RLLIB. Note this is on a per glue basis.

        1. read the training information for the current episode
        2. For each metric in each platform interface in each environment
           update metric container

        Parameters
        ----------
        worker : RolloutWorker
            Reference to the current rollout worker.
        base_env : BaseEnv
            BaseEnv running the episode. The underlying
            env object can be gotten by calling base_env.get_sub_environments().
        policies : dict
            Mapping of policy id to policy objects. In single
            agent mode there will only be a single "default" policy.
        episode : MultiAgentEpisode
            EpisodeV2 object which contains episode
            state. You can use the `episode.user_data` dict to store
            temporary data, and `episode.custom_metrics` to store custom
            metrics for the episode.
        """

        # Don't need to run if we are not evaluating
        if not worker.config.in_evaluation:
            return

        epoch = str(worker.config["training_iteration"]).zfill(EPOCH_ZFILL)
        env = base_env.get_sub_environments()[episode.env_id]
        done_string_list = []
        # Generate the filenames for trajectories
        for platform_name, done_dict in env.done_info.items():
            for done_name, done_state in done_dict.items():
                if done_state:
                    done_string_list.extend([platform_name[0].upper(), done_name, str(done_state)])
        episode_counter_str = str(self._episode_counter).zfill(8)
        worker_index_str = str(worker.env_context.worker_index).zfill(3)
        done_string = "_".join(done_string_list)
        done_string = f"{episode_counter_str}-{worker_index_str}-{done_string}"

        sub_directory = Path(worker._original_kwargs["log_dir"]) / "trajectories" / f"epoch_{epoch}"  # noqa: SLF001
        policy_artifact = {}
        for policy_id in policies:
            policy_artifact.update(
                {
                    policy_id: EpisodeArtifact.PolicyArtifact(
                        preprocessor=get_preprocessor(env.observation_space)(env.observation_space), filters=worker.filters[policy_id]
                    )
                }
            )
        assert isinstance(worker.env, ACT3MultiAgentEnv)
        assert isinstance(worker.env.action_space, gymnasium.spaces.Dict)
        assert isinstance(worker.env.observation_space, gymnasium.spaces.Dict)
        space_definitions = EpisodeArtifact.SpaceDefinitions(
            action_space=worker.env._raw_action_space,  # noqa: SLF001
            normalized_action_space=worker.env.action_space,
            observation_space=worker.env._raw_observation_space,  # noqa: SLF001
            normalized_observation_space=worker.env.observation_space,
        )
        episode_artifact = EpisodeArtifact(
            test_case=episode.user_data["test_case"],
            env_config=episode.user_data["env_config"],
            artifacts_filenames=episode.user_data["episode_artifacts_filenames"],
            start_time=episode.user_data["start_time"],
            duration_sec=(datetime.now() - episode.user_data["start_time"]).total_seconds(),
            steps=episode.user_data["steps"],
            walltime_sec=episode.user_data["walltime_sec"],
            dones=episode.user_data["dones"],
            frame_rate=env.simulator.frame_rate,
            episode_state=episode.user_data["episode_state"],
            simulator_info=episode.user_data["simulator_info"],
            worker_index=worker.env_context.worker_index,
            params=episode.user_data["params"],
            parameter_values=episode.user_data["parameter_values"],
            observation_units=env._observation_units,  # noqa: SLF001
            platform_to_agents=env.platform_to_agents,
            agent_to_platforms=env.agent_to_platforms,
            done_config=episode.user_data["done_config"],
            initial_state=episode.user_data["initial_state"],
            policy_artifact=policy_artifact,
            space_definitions=space_definitions,
            platform_serializer=self.platform_serializer(),
        )

        if not os.path.exists(sub_directory):
            os.makedirs(sub_directory, exist_ok=True)

        with open(sub_directory / f"{done_string}.pickle", "wb") as handle:
            pickle.dump(episode_artifact, handle, protocol=pickle.HIGHEST_PROTOCOL)
        self._episode_counter += 1

    def on_postprocess_trajectory(
        self,
        *,
        worker: RolloutWorker,
        episode: EpisodeV2,
        agent_id: AgentID,
        policy_id: PolicyID,
        policies: dict[PolicyID, Policy],
        postprocessed_batch: SampleBatch,
        original_batches: dict[AgentID, tuple[Policy, SampleBatch]],
        **kwargs,
    ) -> None:
        ...

        # This is an overwrite.
        #
        # Default Evaluation callbacks writes EpisodeArtifact to processed_batch.
        # This creates problems during training when batches need to be concatenated and
        # EpisodeArtifact cannot be concatenated. Only seems to effect single policy
        # where two different platforms contribute to a single Sample Batch



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
"""

import abc
from collections import OrderedDict
from enum import Enum

import gymnasium
from pydantic import BaseModel

from corl.libraries.env_func_base import EnvFuncBase
from corl.libraries.environment_dict import DoneDict
from corl.libraries.state_dict import StateDict
from corl.simulators.base_parts import BaseTimeSensor
from corl.simulators.base_simulator import BaseSimulatorState


class DoneStatusCodes(Enum):
    """reward states for the done conditions"""

    WIN = 1
    PARTIAL_WIN = 2
    DRAW = 3
    PARTIAL_LOSS = 4
    LOSE = 5


class DoneFuncBaseValidator(BaseModel):
    """Initialize the done condition

    All done conditions have three common parameters, with any others being handled by subclass validators

    Parameters
    ----------
    name : str
        A name applied to this done condition, by default the name of the class.
    agent_name : str
        Name of the agent to which this done condition applies, This is only valid
        if the done condition came from an agent
    platform_name : str
        Name of the platform to which this done condition applies
    """

    name: str = ""
    agent_name: str | None = None
    platform_name: str


class DoneFuncBase(EnvFuncBase):
    """Base implementation for done functors"""

    def __init__(self, **kwargs) -> None:
        self.config: DoneFuncBaseValidator = self.get_validator()(**kwargs)
        self.config.name = self.config.name if self.config.name else type(self).__name__

    @staticmethod
    def get_validator() -> type[DoneFuncBaseValidator]:
        """
        get validator for this Done Functor

        Returns:
            DoneFuncBaseValidator -- validator the done functor will use to generate a configuration
        """
        return DoneFuncBaseValidator

    @property
    def agent(self) -> str | None:
        """The agent to which this done is applied"""
        return self.config.agent_name

    @property
    def platform(self) -> str:
        """The platform to which this done is applied"""
        return self.config.platform_name

    @property
    def name(self) -> str:
        return self.config.name

    @staticmethod
    def _get_platform_time(platform):
        if sensor := [s for s in platform.sensors if isinstance(s, BaseTimeSensor)]:
            return sensor[0].get_measurement()[0]

        raise ValueError("Did not find time sensor type (BaseTimeSensor)")

    @abc.abstractmethod
    def __call__(
        self,
        observation: OrderedDict,
        action: OrderedDict,
        next_observation: OrderedDict,
        next_state: BaseSimulatorState,
        observation_space: gymnasium.Space,
        observation_units: OrderedDict,
    ) -> bool:
        ...


class SharedDoneFuncBaseValidator(BaseModel):
    """
    name : str
        The name of this done condition
    """

    name: str = ""


class SharedDoneFuncBase(EnvFuncBase):
    """Base implementation for global done functors

    Global done functors are not associated with a single agent, but the environment as a whole.  They use a modified call syntax that
    receives the done dictionary and done info from the per-agent done functors.
    """

    ALL = "__all__"

    def __init__(self, **kwargs) -> None:
        self.config: SharedDoneFuncBaseValidator = self.get_validator()(**kwargs)
        self.config.name = self.config.name if self.config.name else type(self).__name__

    @staticmethod
    def get_validator() -> type[SharedDoneFuncBaseValidator]:
        """
        gets the validator for this SharedDoneFuncBase

        Returns:
            SharedDoneFuncBaseValidator -- validator for this class
        """
        return SharedDoneFuncBaseValidator

    @property
    def name(self) -> str:
        return self.config.name

    @abc.abstractmethod
    def __call__(
        self,
        observation: OrderedDict,
        action: OrderedDict,
        next_observation: OrderedDict,
        next_state: StateDict,
        observation_space: StateDict,
        observation_units: StateDict,
        local_dones: DoneDict,
        local_done_info: OrderedDict,
    ) -> DoneDict:
        ...



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
"""
import logging
from abc import abstractmethod
from typing import Any, TypeVar

import pandas as pd
from pydantic import field_validator

from corl.episode_parameter_providers.base_network_epp import BaseNetworkParameterProvider, EppUpdate, EppUpdateSignaller
from corl.evaluation.connection.base_eval_connection import BaseEvalConnection
from corl.evaluation.connection.signal import Slot
from corl.evaluation.runners.section_factories.test_cases.test_case_manager import (
    TestCaseIndex,
    TestCaseStrategy,
    TestCaseStrategyValidator,
)
from corl.libraries.context import get_current_context
from corl.libraries.factory import Factory

T = TypeVar("T")


class BaseNetworkTestCaseStrategyValidator(TestCaseStrategyValidator, Factory):
    """
    A validator for the NetworkTestCaseStrategy class
    """

    @field_validator("type", mode="after")
    @classmethod
    def subclass_validator(cls, v, info):
        if not issubclass(v, BaseNetworkParameterProvider):
            raise TypeError(f"Invalid type {v}; must be subclass of {BaseNetworkParameterProvider}")
        return v


class BaseNetworkTestCaseStrategy(TestCaseStrategy, EppUpdateSignaller[T]):
    """
    The concrete implementation of a TestCaseStrategy for the NetworkedParameterProvider.
    """

    def __init__(self, **kwargs) -> None:
        self.config: BaseNetworkTestCaseStrategyValidator
        super().__init__(**kwargs)

        self._logger = logging.getLogger(type(self).__name__)

        self._test_case: pd.DataFrame | list[dict[str, Any]] = [{}]
        self._cur_updates: T | None = None

        connection: BaseEvalConnection | None
        if (connection := get_current_context().get("connection")) is not None:
            connection.modify_epp_signal.register(Slot(self))

    @staticmethod
    def get_validator() -> type[BaseNetworkTestCaseStrategyValidator]:
        """
        Method to return validator for TestCaseStrategy
        """
        return BaseNetworkTestCaseStrategyValidator

    def update_rllib_config(self, config: dict):
        """
        Method responsible for mutating rllib config, overwriting the task's default EPPs with NetworkProvider.
        """
        # Override the environment EPP
        config["env_config"]["episode_parameter_provider"] = {
            "type": f"{self.config.type.__module__}.{self.config.type.__qualname__}",
            "config": {**self.config.config, "registrar": self, "epp_variety": "environment", "agent_id": None},
        }

        # Override the agents EPP
        for agent_id, value in config["env_config"]["agents"].items():
            value.class_config.config["episode_parameter_provider"] = {
                "type": f"{self.config.type.__module__}.{self.config.type.__qualname__}",
                "config": {**self.config.config, "registrar": self, "epp_variety": "agent", "agent_id": agent_id},
            }

        return config

    def get_test_cases(self):
        """
        Method responsible for retrieving test cases planned for evaluation
        """
        return self._test_case

    def get_test_case_index(self, episode_id: int) -> TestCaseIndex:  # noqa: PLR6301
        return TestCaseIndex(episode_id)

    def __call__(self, updates: T):
        self._cur_updates = updates

        self._test_case = self._create_test_case_for_updates(updates)

        self.epp_update_signal(EppUpdate(self.get_num_test_cases() - 1, updates))

    @abstractmethod
    def _create_test_case_for_updates(self, updates: T) -> pd.DataFrame | list[dict[str, Any]]:
        """Adds the most recent test case"""
        ...
